from typing import Optional, Tuple

import mlx.core as mx
import mlx.nn as nn

from sillm.models.base import BaseModel
from sillm.models.args import ModelArgs
from sillm.modules.rope import init_rope
from sillm.modules.switch import SwitchGLU
import sillm.models.llama as llama

class Attention(llama.Attention):
    """
    Multi-head attention module.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        nn.Module.__init__(self)
        self.args = args

        self.n_heads: int = args.n_heads
        self.n_kv_heads: int = args.n_kv_heads

        self.scale = self.args.head_dim ** -0.5
        
        self.wq = nn.Linear(args.dim, args.n_heads * args.head_dim, bias=True)
        self.wk = nn.Linear(args.dim, args.n_kv_heads * args.head_dim, bias=True)
        self.wv = nn.Linear(args.dim, args.n_kv_heads * args.head_dim, bias=True)
        self.wo = nn.Linear(args.n_heads * args.head_dim, args.dim, bias=True)
        
        self.rope = init_rope(args)

class FeedForward(nn.Module):
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        super().__init__()

        self.num_experts = args.num_local_experts
        self.num_experts_per_tok = args.num_experts_per_tok

        self.gate = nn.Linear(args.dim, self.num_experts, bias=False)
        self.switch_mlp = SwitchGLU(args.dim, args.hidden_dim, self.num_experts, bias=False)

    def __call__(self,
                 x: mx.array
                 ) -> mx.array:
        """
        Args:
            x: Input tensor.
        Returns:
            Output tensor.
        """
        gates = self.gate(x)

        k = self.num_experts_per_tok
        inds = mx.stop_gradient(mx.argpartition(-gates, kth=k - 1, axis=-1)[..., :k])
        scores = mx.take_along_axis(gates, inds, axis=-1)
        scores = mx.softmax(scores, axis=-1, precise=True)

        y = self.switch_mlp(x, inds)
        y = (y * scores[..., None]).sum(axis=-2)

        return y

class TransformerBlock(llama.TransformerBlock):
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        nn.Module.__init__(self)

        self.args = args

        self.n_heads = args.n_heads
        self.dim = args.dim

        self.attention = Attention(args=args)
        self.attention_norm = nn.LayerNorm(args.dim, eps=args.rms_norm_eps)
        self.feed_forward = FeedForward(args)
        self.ffn_norm = nn.LayerNorm(args.dim, eps=args.rms_norm_eps)

class Model(llama.Model):
    """
    Phi-MoE model wrapper.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        BaseModel.__init__(self, args)
        self.args = args

        self.n_layers = args.n_layers
        self.vocab_size = args.vocab_size
        
        self.tok_embeddings = nn.Embedding(args.vocab_size, args.dim)
        self.layers = [TransformerBlock(args=args) for _ in range(args.n_layers)]
        self.norm = nn.LayerNorm(args.dim, eps=args.rms_norm_eps)
        self.output = nn.Linear(args.dim, args.vocab_size, bias=True)

    def preprocess_weights(self,
                           weights: dict
                           ) -> dict:
        num_experts = self.args.num_local_experts
        
        for l in range(self.args.n_layers):
            for k in ("w1", "w2", "w3"):
                for n in ("weight", "scales", "biases"):
                    prefix = f"layers.{l}.feed_forward"

                    if f"{prefix}.experts.0.{k}.{n}" not in weights:
                        continue

                    experts_keys = [prefix + f".experts.{i}.{k}.{n}" for i in range(num_experts)]
                    weights[prefix + f".switch_mlp.{k}.{n}"] = mx.stack([weights[key] for key in experts_keys])

                    for key in experts_keys:
                        del weights[key]

        return weights


import logging
import random

import tqdm

import numpy as np
from sklearn.decomposition import PCA

import mlx.core as mx
import mlx.nn as nn
from mlx.utils import tree_unflatten

from sillm.core.llm import LLM

logger = logging.getLogger("sillm")

def save_vectors(fpath: str,
                 vectors: dict,
                 metadata: dict = {}):
    """
    Save control vectors to a file.
    Args:
        fpath: File path.
        vectors: Control vectors.
        metadata: Metadata.
    """
    mx.save_safetensors(fpath, vectors, metadata=metadata)

def load_vectors(fpath: str):
    """
    Load control vectors from a file.
    Args:
        fpath: File path.
    Returns:
        Control vectors.
    """
    return mx.load(fpath, return_metadata=True)

class ContrastDataset:
    """
    Contrast dataset for control vector training.
    """
    def __init__(self,
                 tokenizer,
                 positive: list,
                 negative: list
                 ):
        self.tokenizer = tokenizer
        self._positive = positive
        self._negative = negative
    
    def iterate(self):
        """
        Iterate over contrast pairs.
        """
        pad_id = self.tokenizer.pad_id
        if pad_id is None:
            pad_id = self.tokenizer.eos_id

        while True:
            pos = random.choice(self._positive)
            neg = random.choice(self._negative)

            lengths = [len(pos), len(neg)]
            max_length = max(lengths)

            batch_arr = np.full((2, max(lengths)), pad_id, np.int32)
            batch_arr[0, max_length-lengths[0]:] = pos
            batch_arr[1, max_length-lengths[1]:] = neg

            yield mx.array(batch_arr)
    
class ConversationContrastDataset(ContrastDataset):
    """
    Conversation dataset with positive and negative prompts for control vector training.
    """
    def __init__(self,
                 tokenizer,
                 template,
                 positive: list,
                 negative: list,
                 responses: list
                 ):
        self.tokenizer = tokenizer

        def get_entry(prompt, response):
            # Apply chat template and tokenize
            messages = [
                {"role": "user", "content": prompt},
                {"role": "assistant", "content": response}
            ]
            text = template.apply_chat_template(messages)

            return self.tokenizer.encode(text)

        self._positive = []
        self._negative = []
        for response in responses:
            for prompt in positive:
                pos = get_entry(prompt, response)
                self._positive.append(pos)
            for prompt in negative:
                neg = get_entry(prompt, response)
                self._negative.append(neg)

class SystemContrastDataset(ContrastDataset):
    """
    Conversation dataset with a system message for control vector training.
    """
    def __init__(self,
                 tokenizer,
                 template,
                 positive: list,
                 negative: list,
                 prompts: list,
                 responses: list
                 ):
        self.tokenizer = tokenizer

        if len(positive) == 0:
            positive = [None]
        if len(negative) == 0:
            negative = [None]

        def get_entry(messages):
            # Apply chat template and tokenize
            text = template.apply_chat_template(messages)

            return self.tokenizer.encode(text)

        self._positive = []
        self._negative = []
        for response in responses:
            for prompt in prompts:
                messages = [
                    {"role": "system", "content": ""},
                    {"role": "user", "content": prompt},
                    {"role": "assistant", "content": response}
                ]
                for system in positive:
                    messages[0]["content"] = system
                    if system is None:
                        pos = get_entry(messages[1:])
                    else:
                        pos = get_entry(messages)
                    self._positive.append(pos)
                for system in negative:
                    messages[0]["content"] = system
                    if system is None:
                        neg = get_entry(messages[1:])
                    else:
                        neg = get_entry(messages)
                    self._negative.append(neg)

class ControlModule(nn.Module):
    """
    Control module wrapper.
    """
    @staticmethod
    def from_module(module: nn.Module):
        """
        Create control module wrapper for a given module.
        Args:
            module: Module to wrap.
        Returns:
            Control module wrapper.
        """
        control = ControlModule()
        control.module = module
        control.name = module.name

        return control

    def __init__(self):
        super().__init__()

        self.hidden_states = None

        self.capture = False
        self.mode = 'output'
        self.vector = None
        self.alpha = 1.0
        self.beta = -1.0

    def hook(self, func, x, *args, **kwargs):
        """
        Hook function for forward pass.
        Stores hidden states during training and applies control vector during inference.
        """
        output = func(x, *args, **kwargs)

        if isinstance(output, tuple):
            h = output[0]
        else:
            h = output

        # Store hidden states
        if self.capture:
            if self.mode == "input":
                mx.eval(x)
                self.hidden_states = x
            elif self.mode == "output":
                mx.eval(h)
                self.hidden_states = h

        # Apply control vector
        if self.vector is not None:
            if self.vector.dtype != h.dtype:
                self.vector = self.vector.astype(h.dtype)

            if self.mode == 'output':
                norm_pre = mx.linalg.norm(h, axis=-1, keepdims=True)

                h += self.alpha * self.vector

                norm_post = mx.linalg.norm(h, axis=-1, keepdims=True)
                norm = norm_pre / norm_post
                norm = mx.where(mx.isnan(norm), 1.0, norm)
                
                h *= norm
            elif self.mode == 'input':
                vector = self.vector * self.alpha
                vector_reshaped = mx.reshape(vector, (-1, 1)).astype(h.dtype)
                proj = (h @ vector_reshaped) * vector

                h += proj * self.beta
            else:
                raise ValueError(f"Invalid mode: {self.mode}")
        else:
            return output
            
        if isinstance(output, tuple):
            return (h,) + output[1:]
        else:
            return h
    
    def __call__(self, x, *args, **kwargs):
        return self.hook(self.module, x, *args, **kwargs)
    
    def forward(self, x, *args, **kwargs):
        return self.hook(self.module.forward, x, *args, **kwargs)
    
class ControlledLLM(LLM):
    """
    LLM model control vector wrapper.
    """
    @staticmethod
    def from_model(llm: LLM):
        return ControlledLLM(llm)
    
    def __init__(self, llm: LLM):
        self.model = llm.model
        self.tokenizer = llm.tokenizer
        self.args = llm.args
        self._quantization = llm._quantization
        self._mapping = llm._mapping

        # Set module names
        for key, module in self.model.named_modules():
            module.name = key

        self._mode = None
        self._control_modules = []

    def get_module_index(self,
                         min_layer: int = 0,
                         max_layer: int = 0,
                         segment: str = None,
                         transformer: bool = False,
                         attention: bool = False,
                         feed_forward: bool = False,
                         attention_norm: bool = False,
                         ffn_norm: bool = False,
                         attention_output: bool = False,
                         ffn_down: bool = False
                         ):
        """
        Get the module index for the given layers and module types.
        Args:
            min_layer: Minimum layer index.
            max_layer: Maximum layer index.
            segment: Segment of layers (all/intermediate/core).
            transformer: Include transformer modules.
            attention: Include attention modules.
            feed_forward: Include feed forward modules.
            attention_norm: Include attention normalization modules.
            ffn_norm: Include feed forward normalization modules.
            attention_output: Include attention output modules.
            ffn_down: Include feed forward down modules.
        Returns:
            List of control module indices.
        """
        control_index = []

        if segment == "all":
            min_layer = 0
            max_layer = len(self.model.layers) - 1
        elif segment == "intermediate":
            min_layer = int(len(self.model.layers) * 0.25)
            max_layer = int(len(self.model.layers) * 0.75)
        elif segment == "core":
            min_layer = int(len(self.model.layers) * 0.33)
            max_layer = int(len(self.model.layers) * 0.66)

        if max_layer < 1:
            max_layer = len(self.model.layers) + max_layer

        assert min_layer >= 0 and min_layer <= len(self.model.layers)
        assert max_layer >= 0 and max_layer < len(self.model.layers)

        for i in range(min_layer, max_layer+1):
            if transformer:
                control_index.append(f"layers.{i}")
            if attention:
                control_index.append(f"layers.{i}.attention")
            if feed_forward:
                control_index.append(f"layers.{i}.feed_forward")
            if attention_norm and hasattr(self.model.layers[i], "attention_norm"):
                control_index.append(f"layers.{i}.attention_norm")
            if ffn_norm and hasattr(self.model.layers[i], "ffn_norm"):
                control_index.append(f"layers.{i}.ffn_norm")
            if attention_output and hasattr(self.model.layers[i].attention, "wo"):
                control_index.append(f"layers.{i}.attention.wo")
            if ffn_down and hasattr(self.model.layers[i].feed_forward, "w2"):
                control_index.append(f"layers.{i}.feed_forward.w2")

        return control_index
    
    def init_control(self,
                     control_index: list = [],
                     mode: str = 'output'
                     ):
        """
        Initialize control modules.
        Args:
            control_index: Index of control modules.
        """
        self._mode = mode

        if len(self._control_modules) > 0:
            logger.warning("Control modules already initialized")

        control_modules = {}
        control_tree = []

        for name, module in self.model.named_modules():
            if name in control_index:
                control_module = ControlModule.from_module(module)
                control_module.mode = mode
                control_modules[name] = control_module

                control_tree.append((name, control_module))

        self._control_modules = control_modules

        control_tree = tree_unflatten(control_tree)
        if 'layers' in control_tree and len(control_tree['layers']) != len(self.model.layers):
            control_tree['layers'] += [{} for _ in range(len(self.model.layers) - len(control_tree['layers']))]

        self.model.update_modules(control_tree)
        logger.debug(f"Initialized control modules: {', '.join(control_modules.keys())}")

    def reset(self):
        """
        Unwrap all control modules.
        """
        control_tree = []
        for name, module in self._control_modules.items():
            control_tree.append((name, module.module))
        
        control_tree = tree_unflatten(control_tree)
        if 'layers' in control_tree and len(control_tree['layers']) != len(self.model.layers):
            control_tree['layers'] += [{} for _ in range(len(self.model.layers) - len(control_tree['layers']))]

        self.model.update_modules(control_tree)

        self._control_modules = {}

    def capture(self,
                mode: bool = True
                ):
        """
        Enable or disable capturing module states.
        Args:
            mode: Enable or disable capturing.
        """
        for module in self._control_modules.values():
            module.capture = mode

            if mode is False:
                module.hidden_states = None
    
    def set_control_vectors(self,
                            control_vectors: dict,
                            alpha: float = 1.0,
                            beta: float = -1.0
                            ):
        """
        Set control vectors for control modules.
        Args:
            control_vectors: Dictionary of control vectors.
        """
        if len(self._control_modules) == 0:
            control_index = list(control_vectors.keys())
            self.init_control(control_index)

        for name, vector in control_vectors.items():
            if name in self._control_modules:
                module = self._control_modules[name]

                module.vector = vector
                module.alpha = alpha
                module.beta = beta
            else:
                logger.warning(f"Attempting to set control vector for missing module: {name}")

    def set_control_vector(self,
                           control_vector: mx.array,
                           alpha: float = 1.0,
                           beta: float = -1.0
                           ):
        """
        Set a single control vector for all control modules.
        Args:
            control_vector: Control vector.
            alpha: Control vector coefficient.
        """      
        for module in self._control_modules.values():
            module.vector = control_vector
            module.alpha = alpha
            module.beta = beta

    def save_control_vectors(self,
                             fpath: str
                             ):
        """
        Save control vectors to a file.
        Args:
            fpath: File path.
        """
        vectors = {}
        for name, module in self._control_modules.items():
            vectors[name] = module.vector

        metadata = {
            "mode": self._mode
        }

        save_vectors(fpath, vectors, metadata)

        logger.debug(f"Saved control vectors to: {fpath}")

    def load_control_vectors(self,
                             fpath: str
                             ):
        """
        Load control vectors from a file.
        Args:
            fpath: File path.
        """
        vectors, metadata = load_vectors(fpath)
        control_index = list(vectors.keys())

        self.init_control(control_index=control_index, mode=metadata["mode"])
        self.set_control_vectors(vectors)

        logger.debug(f"Loaded control vectors from: {fpath}")

    def set_coeff(self,
                  alpha: float = 1.0,
                  beta: float = -1.0
                  ):
        """
        Set control vector coefficients for all control modules.
        Args:
            alpha: Vector coefficient.
            beta: Projection coefficient.
        """
        for module in self._control_modules.values():
            module.alpha = alpha
            module.beta = beta

        logger.debug(f"Set control vector coefficients: alpha={alpha}, beta={beta}")

    def modify_weights(self,
                       vectors: dict,
                       alpha: float = 1.0,
                       beta: float = -1.0
                       ):
        """
        Modify model weights with the given direction.
        Args:
            vectors: Modification vectors.
            alpha: Vector coefficient.
            beta: Projection coefficient.
        """
        for name, module in tqdm.tqdm(self.model.named_modules()):
            if name in vectors:
                vector = vectors[name].astype(mx.float32) * alpha
                vector_reshaped = mx.reshape(vector, (-1, 1))

                weight = module.weight.astype(mx.float32)
                proj = (vector @ weight) * vector_reshaped

                module.weight += proj.astype(module.weight.dtype) * beta

        logger.debug(f"Modified weights for modules: {', '.join(vectors.keys())}")

    def train(self,
              dataset: ContrastDataset,
              method: str = "pca_center",
              steps: int = 100,
              stack: bool = False
              ):
        """
        Train control vectors.
        Args:
            dataset: Contrast dataset.
            method: Training method (pca_center, pca_diff, cluster_mean).
            steps: Training steps.
        Returns:
            Dictionary of directions.
        """
        hidden_states_pos = {}
        hidden_states_neg = {}
        if stack:
            hidden_states_pos["stack"] = []
            hidden_states_neg["stack"] = []
        else:
            for name in self._control_modules.keys():
                hidden_states_pos[name] = []
                hidden_states_neg[name] = []

        # Enable training mode to store hidden states
        self.capture(mode=True)

        logger.info("Evaluating forward pass of training data to capture hidden states")
        for _ in tqdm.tqdm(range(steps)):
            batch = next(dataset.iterate())
            self.model(batch)

            for name, module in self._control_modules.items():
                if stack:
                    name = "stack"
                
                hidden_states_pos[name] += module.hidden_states[0].tolist()
                hidden_states_neg[name] += module.hidden_states[1].tolist()

        # Disable training mode
        self.capture(mode=None)

        def project_onto_direction(hidden_states, direction):
            mag = mx.linalg.norm(direction)
            projection = (hidden_states @ direction) / mag

            return projection

        logger.info("Training directions")
        directions = {}
        for name in tqdm.tqdm(hidden_states_pos.keys()):
            h_pos = np.array(hidden_states_pos[name]).astype(np.float32)
            h_neg = np.array(hidden_states_neg[name]).astype(np.float32)

            if method.startswith("pca_"):
                if method == "pca_center":
                    h_center = (h_pos + h_neg) / 2
                    h_pos = h_pos - h_center
                    h_neg = h_neg - h_center
                    h_train = np.vstack([h_pos, h_neg])
                elif method == "pca_diff":
                    h_train = h_pos - h_neg
                    
                pca_model = PCA(n_components=1, whiten=False).fit(h_train)
                dir = pca_model.components_.astype(np.float32).squeeze()
            elif method == "cluster_mean":
                # Simple differences of positive and negative means
                h_pos_mean = h_pos.mean(axis=0, keepdims=True)
                h_neg_mean = h_neg.mean(axis=0, keepdims=True)
                dir = (h_pos_mean - h_neg_mean).squeeze()
                dir = dir / np.linalg.norm(dir)
            dir = mx.array(dir)

            projected_pos = project_onto_direction(h_pos, dir)
            projected_neg = project_onto_direction(h_neg, dir)

            projected_pos_mean = (projected_pos > projected_neg).mean().item()
            projected_neg_mean = (projected_neg > projected_pos).mean().item()
            
            logger.debug(f"Extracted direction for {name}")
            logger.debug(f"Projected positive mean: {projected_pos_mean}")
            logger.debug(f"Projected negative mean: {projected_neg_mean}")
            if projected_neg_mean > projected_pos_mean:
                dir *= -1
                logger.debug(f"Flipped direction for module: {name}")

            directions[name] = dir

        return directions


import logging
import time
import pathlib
import json
import typing

import mlx.core as mx
import mlx.nn as nn
from mlx.utils import tree_flatten, tree_unflatten

from .tokenizer import Tokenizer
import sillm.models as models
import sillm.models.args as args
import sillm.utils.sampling as sampling
from sillm.training.dataset import Dataset
from sillm.core.cache import KVCache, QuantizedKVCache, PromptCache
from sillm.modules.switch import SwitchLinear
from sillm.experimental.logit_filter import LogitFilter

logger = logging.getLogger("sillm")

model_map = {
    "llama":        models.llama.Model,
    "mistral":      models.llama.Model,
    "gemma":        models.gemma.Model,
    "gemma2":       models.gemma2.Model,
    "mixtral":      models.mixtral.Model,   
    "phi":          models.phi.Model,
    "phi3":         models.phi3.Model,
    "phimoe":       models.phimoe.Model,
    "starcoder2":   models.starcoder2.Model,
    "qwen2":        models.qwen2.Model,
    "dbrx":         models.dbrx.Model,
    "cohere":       models.cohere.Model,
    "nemotron":     models.nemotron.Model,
    "pharia-v1":    models.pharia1.Model,
    "granite":      models.granite.Model,
    "mobilellm":    models.llama.Model,
    "exaone":       models.llama.Model,
}

class LLM():
    """
    LLM model wrapper.
    """
    def __init__(self,
                 tokenizer: Tokenizer,
                 args: args.ModelArgs
                 ):
        """
        Args:
            tokenizer: Tokenizer instance.
            args: Model arguments.
        """
        self.args = args
        self._quantization = None
        self._mapping = None

        if args.model_type in model_map:
            model_class = model_map[args.model_type]
            self.model = model_class(args)
        else:
            raise NotImplementedError(f"Model type {args.model_type} is not supported")
        self.model.train(mode=False)

        self._update_names()

        self.tokenizer = tokenizer

    def init_kv_cache(self,
                      step: int = 256,
                      quantized: bool = False,
                      group_size: int = 32,
                      bits: int = 4,
                      ):
        """
        Initialize KV cache for this model type.
        """
        if quantized:
            logger.debug(f"Initializing quantized KV cache ({bits} bits / group size {group_size}) with step size {step}")

            return QuantizedKVCache.for_model(self, step, group_size, bits)
        else:
            logger.debug(f"Initializing KV cache with step size {step}")

            return KVCache.for_model(self, step)

    def init_description(self, model_path):
        """
        Set model description.
        Args:
            description: Model description.
        """
        self.path = pathlib.Path(model_path)
        self.id = self.path.name
        if self.path.is_file():
            self.id = self.path.stem    
        self.created = int(self.path.stat().st_ctime)
    
    def description(self):
        if self.id and self.created:
            return {
                "id": self.id,
                "object": "model",
                "created": self.created
            }
        else:
            raise ValueError("Model is missing an ID or creation time")

    @property
    def size(self):
        """
        Get model size in bytes.
        """
        total_size = 0
        for module in self.model.modules():
            if isinstance(module, nn.QuantizedLinear):
                total_size += int(module.weight.size * 32 / module.bits)
            elif isinstance(module, nn.Linear):
                total_size += module.weight.size

        return total_size
    
    @property
    def max_tokens(self):
        """
        Get maximum number of tokens.
        """
        return self.args.max_position_embeddings
    
    def _update_names(self):
        """
        Update module names.
        """
        for name, module in self.model.named_modules():
            module._name = name
    
    def preprocess_weights(self,
                           weights: dict
                           ) -> dict:
        """
        Preprocess model weights.
        """
        if hasattr(self.model, "preprocess_weights"):
            logger.debug(f"Preprocessing model weights")

            return self.model.preprocess_weights(weights)
        
        return weights

    def update_weights(self,
                       weights: dict,
                       mapping: dict = None
                       ):
        """
        Update model weights.
        Args:
            weights: Weights to update.
        """
        # Update weights
        weights = tree_unflatten(list(weights.items()))
        self.model.update(weights)
        mx.eval(self.model.parameters())

        # Add key mapping for potential quantization
        for key1 in list(mapping.keys()):
            key2 = mapping[key1]
            if key1.endswith(".weight"):
                mapping[key1.removesuffix(".weight") + ".scales"] = key2.removesuffix(".weight") + ".scales"
                mapping[key1.removesuffix(".weight") + ".biases"] = key2.removesuffix(".weight") + ".biases"
        
        # Update key mapping
        self._mapping = mapping

    def verify_weights(self,
                       weights: dict
                       ) -> bool:
        """
        Verify that weights for all modules are present and shapes match.
        Args:
            weights: Weights to verify.
        Returns:
            True if all weights are present and shapes match, False otherwise.
        """
        model_params = tree_flatten(self.model.parameters())
        result = True

        for name, weight in model_params:
            if name not in weights:
                result = False

                logger.warn(f"Key {name} not found in weights")
            elif weight.shape != weights[name].shape:
                result = False

                logger.warn(f"Shape mismatch for key {name}: {weight.shape} != {weights[name].shape}")

        model_keys = {name for name, _ in model_params}
        for name in weights:
            if name not in model_keys:
                logger.debug(f"Unused key {name} in weights")

        return result

    def save_weights(self,
                     weights_path: str
                     ):
        """
        Save model weights into a single safetensors file.
        Args:
            weights_path: Path to weights file.
        """
        state = dict(tree_flatten(self.model.parameters()))
        mx.save_safetensors(weights_path, state)

        logger.info(f"Saved model weights to {weights_path}")

    def save_shards(self,
                    weights_path: str,
                    max_shard_size: int = 5<<30
                    ):
        """
        Save model weights into shards.
        Args:
            weights_dir: Path to weights directory.
        """
        weights_path = pathlib.Path(weights_path)

        state = dict(tree_flatten(self.model.parameters()))

        shards = []
        weight_map = {}
        shard, shard_size = {}, 0
        total_size = 0
        for key, value in state.items():
            if self._mapping is not None:
                key = self._mapping[key]

            if shard_size + value.nbytes > max_shard_size:
                shards.append(shard)
                shard, shard_size = {}, 0

            shard[key] = value
            weight_map[key] = len(shards)

            shard_size += value.nbytes
            total_size += value.nbytes
        shards.append(shard)

        def save_shard(shard, shard_path):
            mx.save_safetensors(shard_path, shard)
            logger.debug(f"Saved shard to {shard_path}")

        if len(shards) > 1:
            for i, shard in enumerate(shards):
                save_shard(shard, str(weights_path / f"model-{i+1:05d}-of-{len(shards):05d}.safetensors"))

            for key, i in weight_map.items():
                weight_map[key] = f"model-{i+1:05d}-of-{len(shards):05d}.safetensors"
        else:
            save_shard(shard, str(weights_path / "model.safetensors"))

            for key in weight_map:
                weight_map[key] = "model.safetensors"

        index_path = str(weights_path / "model.safetensors.index.json")
        with open(index_path, "w") as f:
            index_data = {
                "metadata": {
                    "total_size": total_size,
                },
                "weight_map": weight_map
            }

            f.write(json.dumps(index_data, indent=4))
            logger.debug(f"Saved weight index to {index_path}")

    def save(self,
             model_path: str,
             max_shard_size: int = 5<<30
             ):
        """
        Save model.
        Args:
            model_path: Path to model directory.
        """
        model_path = pathlib.Path(model_path)
        model_path.mkdir(parents=True, exist_ok=True)

        # Save model config
        config_path = model_path / "config.json"
        self.args.save_config(config_path)
        logger.debug(f"Saved model config to {config_path}")

        # Save tokenizer
        self.tokenizer.save(model_path)
        logger.debug(f"Saved tokenizer to {model_path}")

        # Save model weights
        self.save_shards(model_path, max_shard_size=max_shard_size)

        logger.info(f"Saved model to {model_path}")

    def astype(self,
               dtype: str
               ):
        """
        Cast model weights of linear modules to a different data type.
        """
        if dtype not in ("float16", "float32", "bfloat16"):
            raise ValueError(f"Invalid data type {dtype}")
        
        dtype = getattr(mx, dtype)
        for _, module in self.model.named_modules():
            if isinstance(module, nn.Linear) and module.weight.dtype != dtype:
                module.weight = module.weight.astype(dtype)

        logger.info(f"Cast model weights to {dtype}")
        
    def quantize(self,
                 group_size: int = 32,
                 bits: int = 4,
                 weights: dict = None,
                 ):
        """
        Quantize model.
        Args:
            group_size: Group size for quantization.
            bits: Number of bits for quantization.
            weights: Keys to initialize quantization for existing weights.
        """
        if self._quantization is None:
            quantization = {
                "group_size": group_size,
                "bits": bits
            }
            self._quantization = quantization
            self.args.quantization = quantization

            if weights is None:
                class_predicate = lambda p, m: (isinstance(m, (nn.Linear, nn.Embedding, SwitchLinear)) and
                                                ".gate." not in p
                                                )
            else:
                class_predicate = lambda p, m: (isinstance(m, (nn.Linear, nn.Embedding, SwitchLinear)) and
                                                f"{p}.scales" in weights
                                                )
            
            nn.quantize(self.model, class_predicate=class_predicate, **quantization)

            logger.info(f"Quantized model with group size {group_size} and {bits} bits")
        else:
            logger.warn(f"Model is already quantized with group size {group_size} and {bits} bits")

    def dequantize(self):
        """
        Dequantize model.
        """
        if self._quantization is None:
            logger.warn(f"Model is not quantized")
        else:
            layers = []
            for name, module in self.model.named_modules():
                if isinstance(module, nn.QuantizedLinear):
                    bias = "bias" in module
                    weight = mx.dequantize(module.weight, module.scales, module.biases, module.group_size, module.bits).astype(mx.float16)
                    output_dims, input_dims = weight.shape
                    linear = nn.Linear(input_dims, output_dims, bias=bias)
                    linear.weight = weight
                    if bias:
                        linear.bias = module.bias

                    layers.append((name, linear))
                elif isinstance(module, nn.QuantizedEmbedding):
                    weight = mx.dequantize(module.weight, module.scales, module.biases, module.group_size, module.bits).astype(mx.float16)
                    num_embeddings, dims = weight.shape
                    embedding = nn.Embedding(num_embeddings, dims)
                    embedding.weight = weight
                    layers.append((name, embedding))
            
            if len(layers) > 0:
                self.model.update_modules(tree_unflatten(layers))

            self._quantization = None
            self.args.quantization = None

            logger.info(f"Dequantized model")

    def perplexity(self,
                   dataset: Dataset,
                   batch_size: int = 4
                   ):
        """
        Calculate perplexity for an input text.
        Args:
            text: Input text.
        """
        for _, batch in zip(
            range(len(dataset)),
            dataset.iterate_batches(batch_size),
        ):
            losses, _, _ = self.model.loss(*batch)
        
            yield mx.exp(mx.mean(losses)).item()

    def generate(self, *args, **kwargs):
        """
        Iterator for generating tokens.
        Args:
            prompt: Prompt to start generation.
            temperature: Sampling temperature.
            max_tokens: Max number of tokens to generate.
            repetition_penalty: Repetition penalty.
            repetition_window: Repetition window.
            logprobs: Return logprobs.
            token_ids: Return token IDs.
            flush: Flush buffer every n tokens.
            extra_stop_tokens: Additional stop tokens.
        Yields:
            Tuple of generated text and metadata.
        """
        yield from generate(self.model, self.tokenizer, *args, **kwargs)

    def completion(self, *args, **kwargs) -> str:
        """
        Generate a completion and wait for all tokens.
        Args:
            prompt: Prompt to start generation.
            temperature: Sampling temperature.
            max_tokens: Max number of tokens to generate.
            repetition_penalty: Repetition penalty.
            repetition_window: Repetition window.
            logprobs: Return logprobs.
            token_ids: Return token IDs.
            flush: Flush buffer every n tokens.
            extra_stop_tokens: Additional stop tokens.
        Returns:
            Tuple of generated text and metadata.
        """
        result = ""
        for text, metadata in generate(self.model, self.tokenizer, *args, **kwargs):
            result += text

        return result, metadata

def generate(model,
             tokenizer: Tokenizer,
             prompt: str | list | mx.array,
             cache: KVCache = None,
             max_tokens: int = 2048,
             temperature: float = 0.0,
             top_k: int = 0,
             top_p: float = 1.0,
             top_nsigma: float = 0.0,
             repetition_penalty: float = None,
             repetition_window: int = 25,
             logprobs: bool = False,
             token_ids: bool = False,
             flush: int = 5,
             extra_stop_tokens: list = None,
             prompt_cache: PromptCache = None,
             logit_filter: LogitFilter = None
             ):
    start = time.perf_counter()

    # Pre-process inputs
    if isinstance(prompt, str):
        inputs = mx.array(tokenizer.encode(prompt))
    elif isinstance(prompt, list):
        inputs = mx.array(prompt)
    elif isinstance(prompt, mx.array):
        inputs = prompt
    else:
        raise ValueError("Prompt must be a string, list of tokens, or MX array")
    
    # Initialize metadata
    timing = {
        "runtime": 0.0,
        "eval_time": 0.0,
        "tokenizer_time": time.perf_counter() - start
    }
    usage = {
        "prompt_tokens": len(inputs),
        "completion_tokens": 0,
        "total_tokens": 0
    }
    metadata = {
        "timing": timing,
        "usage": usage,
        "logprobs": [],
        "token_ids": [],
        "finish_reason": "length"
    }

    # Define stop tokens
    stop_tokens = tokenizer.special_ids
    
    # Add user-defined stop tokens
    if extra_stop_tokens is not None:
        for token in extra_stop_tokens:
            if isinstance(token, str):
                token_ids = tokenizer.encode(token, bos=False)

                if len(token_ids) > 1:
                    logger.warning(f"Extra stop token '{token}' tokenizes to multiple tokens")
                stop_tokens.add(token_ids[0])
            else:
                stop_tokens.add(token)

    # Initialize token and string buffers
    tokens, text = [], ""

    def sample(logits):
        if temperature == 0:
            y = mx.argmax(logits, axis=-1)
        else:
            # Apply temperature
            logits = logits * (1 / temperature)

            # Apply structure enforcer
            if logit_filter is not None:
                logits = logit_filter(logits)
            # Apply repetition penalty
            if len(tokens) > 0 and repetition_penalty is not None:
                logits = sampling.apply_repetition_penalty(logits, tokens, repetition_penalty=repetition_penalty, repetition_window=repetition_window)
            # Apply top-k sampling
            if top_k > 0:
                logits = sampling.top_k(logits, k=top_k)
            elif 0.0 < top_p < 1.0:
                logits = sampling.top_p(logits, p=top_p)
            elif top_nsigma > 0.0:
                logits = sampling.top_nsigma(logits, n=top_nsigma)
    
            y = mx.random.categorical(logits)

        p = 0.0
        if logprobs:
            p = nn.log_softmax(logits, axis=-1)[0,y].item()

        return y, p

    def generate_step(model, inputs):
        nonlocal cache
        logits = None

        if prompt_cache is not None:
            # Retrieve cached logits and KV cache
            logits, cache = prompt_cache.get(inputs)
        
        if cache is None:
            # Initialize KV cache
            cache = KVCache.for_model(model)

        if logits is None:
            # Initial forward pass through model
            logits = model(inputs[None], cache=cache)
            logits = logits[:, -1, :]

            # Store logits and KV cache in prompt cache
            if prompt_cache is not None:
                prompt_cache.put(inputs, logits, cache)

        y, p = sample(logits)
        yield y, p

        while True:
            # Iterative forward pass through model
            logits = model(y[None], cache=cache)
            logits = logits[:, -1, :]

            y, p = sample(logits)
            yield y, p

    # Main generation loop
    for (token,p), i in zip(generate_step(model, inputs), range(max_tokens)):
        if i == 0:
            mx.eval(token)
            timing["eval_time"] = time.perf_counter() - start

        if token.item() in stop_tokens:
            metadata["finish_reason"] = "stop"
            break

        tokens.append(token.item())

        if logprobs:
            metadata["logprobs"].append(p)

        if (len(tokens) % flush) == 0:
            mx.eval(tokens)

            text_offset = len(text)
            text = tokenizer.decode(tokens)

            timing["runtime"] = time.perf_counter() - start
            usage["completion_tokens"] = len(tokens)
            usage["total_tokens"] = usage["prompt_tokens"] + usage["completion_tokens"]
            if token_ids:
                metadata["token_ids"] += tokens

            yield text[text_offset:], metadata

    mx.async_eval(tokens)

    text_offset = len(text)
    text = tokenizer.decode(tokens)

    timing["runtime"] = time.perf_counter() - start
    usage["completion_tokens"] = i+1
    usage["total_tokens"] = usage["prompt_tokens"] + usage["completion_tokens"]
    if token_ids:
        metadata["token_ids"] = tokens

    yield text[text_offset:], metadata


import collections
import hashlib
import logging

from typing import Optional

import mlx.core as mx
from mlx.utils import tree_map

logger = logging.getLogger("sillm")

########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/fc93c557238e9441835afe2748fd170016cb068b/llms/mlx_lm/models/base.py#L11
########
class KVCache:
    @staticmethod
    def for_model(model,
                  step : int = 256
                  ):
        return [KVCache(step) for _ in range(model.args.n_layers)]

    def __init__(self,
                 step : int = 256
                 ):
        self.keys = None
        self.values = None
        self.offset = 0
        self.step = step

    def update_and_fetch(self,
                         keys,
                         values
                         ):
        B, n_kv_heads, num_steps, k_head_dim = keys.shape
        v_head_dim = values.shape[-1]
        prev_offset = self.offset
        
        if self.keys is None or (prev_offset + num_steps) > self.keys.shape[2]:
            new_steps = (self.step + num_steps - 1) // self.step

            k_shape = (B, n_kv_heads, new_steps * self.step, k_head_dim)
            v_shape = (B, n_kv_heads, new_steps * self.step, v_head_dim)

            new_k = mx.zeros(k_shape, keys.dtype)
            new_v = mx.zeros(v_shape, values.dtype)
            
            if self.keys is not None:
                if prev_offset % self.step != 0:
                    self.keys = self.keys[..., :prev_offset, :]
                    self.values = self.values[..., :prev_offset, :]
                self.keys = mx.concatenate([self.keys, new_k], axis=2)
                self.values = mx.concatenate([self.values, new_v], axis=2)
            else:
                self.keys, self.values = new_k, new_v

        self.offset += num_steps

        self.keys[..., prev_offset : self.offset, :] = keys
        self.values[..., prev_offset : self.offset, :] = values

        return self.keys[..., : self.offset, :], self.values[..., : self.offset, :]
    
    def copy(self):
        new = KVCache(self.step)
        
        if self.keys is not None:
            new.keys = mx.array(self.keys)
        if self.values is not None:
            new.values = mx.array(self.values)
        new.offset = self.offset

        return new

    @property
    def state(self):
        return self.keys, self.values
    
    def quantize(self,
                 group_size: int = 32,
                 bits: int = 4
                 ):
        """
        Quantize the KV cache.
        
        """
        cache = QuantizedKVCache(self.step, group_size, bits)
        cache.offset = self.offset
        if self.keys is not None:
            cache.keys = mx.quantize(self.keys, group_size=group_size, bits=bits)
        if self.values is not None:
            cache.values = mx.quantize(self.values, group_size=group_size, bits=bits)

        return cache

class QuantizedKVCache(KVCache):
    @staticmethod
    def for_model(model,
                  step : int = 256,
                  group_size: int = 32,
                  bits: int = 4,
                  ):
        return [QuantizedKVCache(step, group_size, bits) for _ in range(model.args.n_layers)]
    
    def __init__(self,
                 step : int = 256,
                 group_size: int = 32,
                 bits: int = 4,
                 ):
        self.keys = None
        self.values = None
        self.offset = 0
        self.step = step

        self.group_size = group_size
        self.bits = bits

    def update_and_fetch(self,
                         keys,
                         values
                         ):
        B, n_kv_heads, num_steps, k_head_dim = keys.shape
        v_head_dim = values.shape[-1]
        prev_offset = self.offset
        
        if self.keys is None or (prev_offset + num_steps) > self.keys[0].shape[2]:
            new_steps = (self.step + num_steps - 1) // self.step

            if self.keys is None:
                k_shape = (B, n_kv_heads, new_steps * self.step, k_head_dim)
                v_shape = (B, n_kv_heads, new_steps * self.step, v_head_dim)

                self.keys = mx.quantize(mx.zeros(k_shape, keys.dtype), group_size=self.group_size, bits=self.bits)
                self.values = mx.quantize(mx.zeros(v_shape, keys.dtype), group_size=self.group_size, bits=self.bits)
            else:
                if prev_offset % self.step != 0:
                    self.keys, self.values = tree_map(lambda x: x[..., :prev_offset, :], (self.keys, self.values))

                def expand(x, shape):
                    new_x = mx.zeros((*shape, x.shape[-1]), dtype=x.dtype)
                    return mx.concatenate([x, new_x], axis=-2)

                shape = (B, n_kv_heads, new_steps * self.step)
                self.keys, self.values = tree_map(lambda x: expand(x, shape), (self.keys, self.values))

        self.offset += num_steps

        keys = mx.quantize(keys, group_size=self.group_size, bits=self.bits)
        values = mx.quantize(values, group_size=self.group_size, bits=self.bits)

        for i in range(len(self.keys)):
            self.keys[i][..., prev_offset : self.offset, :] = keys[i]
            self.values[i][..., prev_offset : self.offset, :] = values[i]

        return tree_map(lambda x: x[..., : self.offset, :], (self.keys, self.values))

class PromptCache():
    """
    Cache for prompt logits and KV cache.
    """
    def __init__(self,
                 max_size: int = 10
                 ):
        self.logits = {}
        self.kv_cache = {}

        self.lru = collections.OrderedDict()

        self.max_size = max_size

        # TODO make multi-thread safe?

    def _key(self,
             inputs: mx.array
             ):
        return hashlib.sha256(inputs).hexdigest()
    
    def put(self,
            inputs: mx.array,
            logits: mx.array,
            kv_cache: Optional[KVCache] = None,
            ):
        """
        Add cache entry.
        """
        key = self._key(inputs)

        logger.debug(f"Adding prompt cache entry for key {key}")

        if key not in self.lru:
            self.logits[key] = logits
            self.kv_cache[key] = [c.copy() for c in kv_cache]
            self.lru[key] = 0

        if len(self.lru) > self.max_size:
            pop_key, _ = self.lru.popitem(last=False)
            self.logits.pop(pop_key)
            self.kv_cache.pop(pop_key)

    def get(self,
            inputs: mx.array|str
            ):
        """
        Get cache entry.
        """
        if type(inputs) is mx.array:
            key = self._key(inputs)
        else:
            key = inputs

        if key in self.lru:
            # Update LRU
            self.lru.move_to_end(key)
            self.lru[key] += 1

            logits = self.logits[key]
            kv_cache = [c.copy() for c in self.kv_cache[key]]

            logger.debug(f"Retrieving prompt cache entry for key {key}")
            
            return logits, kv_cache
        
        return None, None


from typing import Optional, Tuple

import numpy as np

import mlx.core as mx
import mlx.nn as nn

from sillm.models.base import BaseModel
from sillm.models.args import ModelArgs
from sillm.modules.switch import SwitchGLU
import sillm.models.llama as llama

########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/f530f56df2738a54982c4541189a8c8d7cd94c44/llms/mlx_lm/models/mixtral.py#L97
########
class FeedForward(nn.Module):
    """
    MoE Feed-forward module for Mixtral models.
    """
    def __init__(self, args: ModelArgs):
        super().__init__()

        self.num_experts = args.moe["num_experts"]
        self.num_experts_per_tok = args.moe["num_experts_per_tok"]

        self.gate = nn.Linear(args.dim, self.num_experts, bias=False)
        self.switch_mlp = SwitchGLU(args.dim, args.hidden_dim, self.num_experts, bias=False)

    def __call__(self,
                 x: mx.array
                 ) -> mx.array:
        """
        Args:
            x: Input tensor.
        Returns:
            Output tensor.
        """
        gates = self.gate(x)

        k = self.num_experts_per_tok
        inds = mx.stop_gradient(mx.argpartition(-gates, kth=k - 1, axis=-1)[..., :k])
        scores = mx.take_along_axis(gates, inds, axis=-1)
        scores = mx.softmax(scores, axis=-1, precise=True)

        y = self.switch_mlp(x, inds)
        y = (y * scores[..., None]).sum(axis=-2)

        return y
    
########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/d8680a89f986492dbc27c36af3294034db26458f/llms/mixtral/mixtral.py#L163
########
class TransformerBlock(nn.Module):
    """
    Transformer block for Mixtral models.
    """
    def __init__(self, args: ModelArgs):
        super().__init__()

        self.n_heads = args.n_heads
        self.dim = args.dim
        self.attention = llama.Attention(args=args)
        self.feed_forward = FeedForward(args=args)
        self.attention_norm = nn.RMSNorm(args.dim, eps=args.norm_eps)
        self.ffn_norm = nn.RMSNorm(args.dim, eps=args.norm_eps)
        self.args = args

    def forward(self,
                 x: mx.array,
                 mask: mx.array = None,
                 cache = None
                 ) -> mx.array:
        r = self.attention(self.attention_norm(x), mask, cache)
        h = x + r
        r = self.feed_forward(self.ffn_norm(h))
        out = h + r

        return out
    
########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/e74889d0fa0fb49d95bfdf6a1dcad907713eb50e/llms/mixtral/mixtral.py#L187
########
class Model(BaseModel):
    """
    Mixtral model wrapper.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        super().__init__(args)
        self.args = args

        self.n_layers = args.n_layers
        self.vocab_size = args.vocab_size
        self.router_aux_loss_coef = args.router_aux_loss_coef

        self.tok_embeddings = nn.Embedding(args.vocab_size, args.dim)
        self.layers = [TransformerBlock(args=args) for _ in range(args.n_layers)]
        self.norm = nn.RMSNorm(args.dim, eps=args.norm_eps)
        self.output = nn.Linear(args.dim, args.vocab_size, bias=False)

    def __call__(self,
                 inputs: mx.array,
                 cache = None
                 ):
        """
        Args:
            inputs: Input tokens.
            cache: Cache from previous forward pass.
        Returns:
            Output logits and cache.
        """
        h = self.tok_embeddings(inputs)

        mask = None
        T = h.shape[1]
        if T > 1:
            mask = nn.MultiHeadAttention.create_additive_causal_mask(T)
            mask = mask.astype(h.dtype)

        if cache is None:
            cache = [None] * len(self.layers)

        for e, layer in enumerate(self.layers):
            h = layer.forward(h, mask, cache[e])

        return self.output(self.norm(h))
    
    def preprocess_weights(self,
                           weights: dict
                           ) -> dict:
        num_experts = self.args.moe["num_experts"]
        
        for l in range(self.args.n_layers):
            for k in ("w1", "w2", "w3"):
                for n in ("weight", "scales", "biases"):
                    prefix = f"layers.{l}.feed_forward"

                    if f"{prefix}.experts.0.{k}.{n}" not in weights:
                        continue

                    experts_keys = [prefix + f".experts.{i}.{k}.{n}" for i in range(num_experts)]
                    weights[prefix + f".switch_mlp.{k}.{n}"] = mx.stack([weights[key] for key in experts_keys])

                    for key in experts_keys:
                        del weights[key]

        return weights


import argparse
import pathlib
import json

def load_chat_template(fpath: str) -> str:
    with open(fpath) as f:
        config = json.loads(f.read())
        
        if "chat_template" in config:
            return config["chat_template"].replace("\n", "")
    
    return None

if __name__ == "__main__":
    # Parse commandline arguments
    parser = argparse.ArgumentParser(description="Extractor for Jinja2 templates from tokenizer config files.")
    parser.add_argument("config", type=str, help="Path to tokenizer config file")
    parser.add_argument("output", type=str, help="Path to output file")
    args = parser.parse_args()

    template = load_chat_template(args.config)

    if template is not None:
        with open(args.output, "w") as f:
            f.write(template)


from sillm.utils.metal import set_memory_limit
from sillm.utils.common import seed, load_yaml
from sillm.utils.log import init_logger, log_arguments, log_memory_usage
from sillm.utils.plot import Plot


import argparse

import sillm
import sillm.utils as utils
from sillm.experimental.distillation import DistillationLoRA

if __name__ == "__main__":
    # Parse commandline arguments
    parser = argparse.ArgumentParser(description="A simple CLI for distilling a model with LoRA.")
    parser.add_argument("draft", type=str, help="The input model directory or file")
    parser.add_argument("target", type=str, help="The output model directory or file")
    parser.add_argument("-c", "--config", default=None, type=str, help="Load YAML configuration file for training")
    parser.add_argument("-t", "--train", default=None, type=str, help="Train the model with training dataset in the file/directory")
    parser.add_argument("-s", "--seed", type=int, default=-1, help="Seed for randomization")
    parser.add_argument("--max_entries", default=None, type=int, help="Max number of entries to load from training dataset (default: unlimited)")
    parser.add_argument("--max_length", default=1024, type=int, help="Max token length per training dataset entry (default: 1024)")
    parser.add_argument("--template", type=str, default=None, help="Chat template (chatml, llama2, alpaca, etc.)")
    parser.add_argument("--lora_layers", default=0, type=int, help="Layers to use for LoRA (default: 0 for all layers)")
    parser.add_argument("--lora_modules", default="query_value", type=str, help="Target modules to use for LoRA: query_value, all_linear")
    parser.add_argument("--lora_rank", default=8, type=int, help="Rank to use for LoRA (default: 8)")
    parser.add_argument("--lora_alpha", default=16.0, type=float, help="Alpha to use for LoRA (default: 16.0)")
    parser.add_argument("--lora_stabilize", default=True, action="store_true", help="Use rank stabilization scaling factor for LoRA layers")
    parser.add_argument("--lora_dropout", default=0.0, type=int, help="Dropout to use for LoRA (default: 0.0)")
    parser.add_argument("--optimizer", type=str, default="adam", help="Optimizer type (default: adam)")
    parser.add_argument("--loss_alpha", default=0.5, type=float, help="Distillation loss alpha (default: 0.5)")
    parser.add_argument("--grad_checkpoint", default=False, action="store_true", help="Use gradient checkpointing")
    parser.add_argument("--grad_accu_steps", type=int, default=1, help="Gradient accumulation steps (default: 1)")
    parser.add_argument("--learning_rate", default=1e-5, type=float, help="Learning rate (default: 1e-5)")
    parser.add_argument("--learning_decay", default=0.0, type=float, help="Learning decay for optimizer schedule (default: 0.0)")
    parser.add_argument("--learning_warmup", default=0, type=int, help="Learning warmup for optimizer schedule (default: 0)")
    parser.add_argument("--epochs", default=1, type=int, help="Number of epochs (default: 1)")
    parser.add_argument("--iterations", default=0, type=int, help="Number of iterations per epoch (default: dataset size)")
    parser.add_argument("--batch_size", default=1, type=int, help="Size of training batches (default: 4)")
    parser.add_argument("--report_steps", default=10, type=int, help="Number of batch iterations per training report (default: 10)")
    parser.add_argument("--eval_steps", default=100, type=int, help="Number of batch iterations per evaluation (default: 100)")
    parser.add_argument("--validation_samples", default=40, type=int, help="Number of validation_samples (default: 40)")
    parser.add_argument("--memory_limit", default=None, type=float, help="Memory limit for training (default: None)")
    parser.add_argument("--relax_memory_limit", default=False, action="store_true", help="Relax memory limit for training")
    parser.add_argument("-v", "--verbose", default=1, action="count", help="Increase output verbosity")
    args = parser.parse_args()

    # Load YAML configuration file
    if args.config is not None:
        utils.load_yaml(args.config, args)

    # Initialize logging
    log_level = 40 - (10 * args.verbose) if args.verbose > 0 else 0
    logger = utils.init_logger(log_level)

    # Log commandline arguments
    if log_level <= 10:
        utils.log_arguments(args.__dict__)

    # Set memory limit
    if args.memory_limit is not None:
        utils.set_memory_limit(args.memory_limit, relaxed=args.relax_memory_limit)

    # Set random seed
    if args.seed >= 0:
        utils.seed(args.seed)

    # Load models
    draft_model = sillm.load(args.draft)
    target_model = sillm.load(args.target)

    # Freeze draft model
    draft_model.model.freeze()

    # Initialize trainable model
    distillation_config = {
        "loss_alpha": args.loss_alpha
    }
    target_model = DistillationLoRA.from_model(target_model, draft_model, **distillation_config)
    
    # Initialize LoRA layers
    lora_config = {
        "num_layers":       args.lora_layers,
        "target_modules":   args.lora_modules,
        "rank":             args.lora_rank,
        "alpha":            args.lora_alpha,
        "stabilize":        args.lora_stabilize,
        "dropout":          args.lora_dropout,
    }
    target_model.init_lora(**lora_config)

    # Set conversation template
    template = sillm.init_template(draft_model.tokenizer, draft_model.args, template_name=args.template)
    
    if args.train is not None:
        # Load training dataset
        dataset_config = {
            "template": template,
            "max_entries": args.max_entries,
            "max_length": args.max_length
        }
        dataset_training, dataset_validation, dataset_test = sillm.load_dataset(draft_model.tokenizer, args.train, **dataset_config)

        # Model training
        training_config = {
            "batch_size":                   args.batch_size,
            "optimizer_type":               args.optimizer,
            "learning_rate":                args.learning_rate,
            "learning_decay":               args.learning_decay,
            "learning_warmup":              args.learning_warmup,
            "gradient_checkpointing":       args.grad_checkpoint,
            "gradient_accumulation_steps":  args.grad_accu_steps,
            "epochs":                       args.epochs,
            "iterations":                   args.iterations,
            "report_steps":                 args.report_steps,
            "eval_steps":                   args.eval_steps,
            "validation_samples":           args.validation_samples,
        }
        target_model.train(dataset_training,
                           dataset_validation,
                           dataset_test,
                           **training_config)


import logging

from functools import partial

import mlx.core as mx
import mlx.nn as nn

from sillm.models.args import ModelArgs

logger = logging.getLogger("sillm")

@partial(mx.compile, shapeless=True)
def relu2(x):
    return nn.relu(x).square()

def init_act(args: ModelArgs):
    if args.hidden_act is None:
        logger.debug("No hidden activation specified. Using SiLU.")
        return nn.SiLU()

    if args.hidden_act == "silu":
        return nn.silu
    elif args.hidden_act == "gelu":
        return nn.gelu
    elif args.hidden_act == "gelu_new":
        return nn.gelu
    elif args.hidden_act == "gelu_pytorch_tanh":
        return nn.gelu
    elif args.hidden_act == "relu2":
        return relu2
    
    logger.warning(f"Unknown hidden activation: {args.hidden_act}. Using SiLU.")
    return nn.SiLU()


from typing import Optional, Tuple

import mlx.core as mx
import mlx.nn as nn

from sillm.models.base import BaseModel, scaled_dot_product_attention
from sillm.core.cache import KVCache
from sillm.models.args import ModelArgs
from sillm.modules.norm import RMSNorm
import sillm.models.llama as llama
    
class Attention(llama.Attention):
    """
    Multi-head attention module.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        super().__init__(args)

        self.attn_logit_softcapping = args.attn_logit_softcapping
        self.scale = 1.0 / (args.query_pre_attn_scalar**0.5)

    def __call__(self,
                 x: mx.array,
                 mask: Optional[mx.array] = None,
                 cache: Optional[KVCache] = None,
                 ) -> mx.array:
        B, L, _ = x.shape

        queries, keys, values = self.wq(x), self.wk(x), self.wv(x)

        queries = queries.reshape(B, L, self.n_heads, -1).transpose(0, 2, 1, 3)
        keys = keys.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)
        values = values.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)

        if cache is not None:
            if cache.offset > 0 and L > 1:
                mask = BaseModel.create_additive_causal_mask(L, offset=cache.offset, dtype=queries.dtype)
                
            queries = self.rope(queries, offset=cache.offset)
            keys = self.rope(keys, offset=cache.offset)
            keys, values = cache.update_and_fetch(keys, values)
        else:
            queries = self.rope(queries)
            keys = self.rope(keys)

        output = scaled_dot_product_attention(queries, keys, values, cache=cache, scale=self.scale, mask=mask)

        ########
        # Attention softcapping
        # Reference:
        # https://github.com/huggingface/transformers/blob/b7ee1e80b912c6cdd93b54dd77af061fde151d28/src/transformers/models/gemma2/modeling_gemma2.py#L259
        ########
        output = output / self.attn_logit_softcapping
        output = mx.tanh(output)
        output = output * self.attn_logit_softcapping

        output = output.transpose(0, 2, 1, 3).reshape(B, L, -1)

        return self.wo(output)

class TransformerBlock(llama.TransformerBlock):
    """
    Transformer block.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        nn.Module.__init__(self)
        self.args = args

        self.n_heads = args.n_heads
        self.dim = args.dim
        
        self.attention = Attention(args=args)
        self.feed_forward = llama.FeedForward(args=args)
        self.attention_norm = RMSNorm(args.dim, eps=args.norm_eps)
        self.ffn_norm = RMSNorm(args.dim, eps=args.norm_eps)
        self.pre_feedforward_layernorm = RMSNorm(args.dim, eps=args.norm_eps)
        self.post_feedforward_layernorm = RMSNorm(args.dim, eps=args.norm_eps)

    def forward(
            self,
            x: mx.array,
            mask: Optional[mx.array] = None,
            cache: Optional[KVCache] = None,
            ) -> mx.array:
        """
        Args:
            x: Input tensor.
            mask: Mask tensor.
            cache: Cache from previous forward pass.
        Returns:
            Output tensor and cache.
        """
        h = self.attention(self.attention_norm(x), mask, cache)
        r = self.ffn_norm(h) + x

        h = self.feed_forward(self.pre_feedforward_layernorm(r))
        out = self.post_feedforward_layernorm(h) + r
        
        return out

########
# References:
# https://github.com/huggingface/transformers/blob/main/src/transformers/models/gemma2/modeling_gemma2.py
########
class Model(llama.Model):
    """
    Gemma model wrapper.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        BaseModel.__init__(self, args)
        self.args = args

        self.n_layers = args.n_layers
        self.vocab_size = args.vocab_size
        self.final_logit_softcapping = args.final_logit_softcapping
        
        self.tok_embeddings = nn.Embedding(args.vocab_size, args.dim)
        self.layers = [TransformerBlock(args=args) for _ in range(args.n_layers)]
        self.norm = RMSNorm(args.dim, eps=args.norm_eps)

    def __call__(self,
                 inputs: mx.array,
                 cache = None
                 ):
        """
        Args:
            inputs: Input tokens.
            cache: Cache from previous forward pass.
        Returns:
            Output logits and cache.
        """
        h = self.tok_embeddings(inputs)
        h = h * (self.args.dim**0.5)

        mask = None
        if h.shape[1] > 1:
            mask = nn.MultiHeadAttention.create_additive_causal_mask(h.shape[1])
            mask = mask.astype(h.dtype)

        if cache is None:
            cache = [None] * len(self.layers)

        for e, layer in enumerate(self.layers):
            h = layer.forward(h, mask, cache[e])

        out = self.tok_embeddings.as_linear(self.norm(h))

        ########
        # Final logit softcapping
        # Reference:
        # https://github.com/huggingface/transformers/blob/b7ee1e80b912c6cdd93b54dd77af061fde151d28/src/transformers/models/gemma2/modeling_gemma2.py#L1083
        ########
        out = out / self.args.final_logit_softcapping
        out = mx.tanh(out)
        out = out * self.args.final_logit_softcapping

        return out


import math

import mlx.core as mx
import mlx.nn as nn
from mlx.utils import tree_map

from sillm.models.args import ModelArgs
from sillm.core.cache import KVCache, QuantizedKVCache

class BaseModel(nn.Module):
    """
    Base class for LLM models.
    """
    def __init__(self, args: ModelArgs):
        super().__init__()
    
    def __call__(self,
                 inputs: mx.array,
                 cache = None
                 ):
        raise NotImplementedError("Class model.Model is used for inheritance only")
    
    ########
    # Based on mlx-examples:
    # https://github.com/ml-explore/mlx-examples/blob/047d4650c4f63d55e5bfbaf8f589c1679cbdd971/lora/lora.py#L151
    ########
    def loss(self,
             inputs: mx.array,
             targets: mx.array,
             loss_masks: mx.array
             ):
        """
        Calculate cross-entropy loss.
        Args:
            inputs: Input tokens.
            targets: Target tokens.
            lengths: Lengths of inputs.
        Returns:
            Cross-entropy loss.
        """
        # Run model on inputs
        logits = self.__call__(inputs)
        logits = logits.astype(mx.float32)

        # Calculate the loss
        cross_entropy_loss = nn.losses.cross_entropy(logits, targets) * loss_masks
        num_tokens = loss_masks.sum()
        loss_value = cross_entropy_loss.sum() / num_tokens

        return loss_value, None, num_tokens

    @staticmethod    
    def create_additive_causal_mask(N: int,
                                    offset: int = 0,
                                    dtype: mx.Dtype = mx.float32
                                    ):
        rinds = mx.arange(offset + N)
        linds = mx.arange(offset, offset + N) if offset else rinds
        mask = linds[:, None] < rinds[None]
        mask = mask.astype(dtype) * mx.finfo(dtype).min
        
        return mask
    
########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/dfa4dd6c93c4c2f81bfed6becb8af5cc3a89ae61/llms/mlx_lm/models/base.py#L56
########
def quantized_scaled_dot_product_attention(queries: mx.array,
                                           keys: mx.array,
                                           values: mx.array,
                                           cache: QuantizedKVCache = None,
                                           scale: float = 1.0,
                                           mask: mx.array = None
                                           ) -> mx.array:
    """
    Scaled dot-product attention with quantized KV cache.
    """
    B, n_q_heads, L, D = queries.shape
    n_kv_heads = keys[0].shape[-3]
    n_repeats = n_q_heads // n_kv_heads
    queries *= scale

    if n_repeats > 1:
        queries = mx.reshape(queries, (B, n_kv_heads, n_repeats, L, D))
        keys = tree_map(lambda x: mx.expand_dims(x, axis=-3), keys)
        values = tree_map(lambda x: mx.expand_dims(x, axis=-3), values)
    scores = mx.quantized_matmul(queries, *keys, transpose=True, group_size=cache.group_size, bits=cache.bits)

    if mask is not None:
        scores += mask
    scores = mx.softmax(scores, axis=-1, precise=True)
    out = mx.quantized_matmul(scores, *values, transpose=False, group_size=cache.group_size, bits=cache.bits)
    if n_repeats > 1:
        out = mx.reshape(out, (B, n_q_heads, L, D))

    return out

def scaled_dot_product_attention(queries: mx.array,
                                 keys: mx.array,
                                 values: mx.array,
                                 cache: KVCache = None,
                                 scale: float = 1.0,
                                 mask: mx.array = None
                                 ) -> mx.array:
    """
    Scaled dot-product attention that transparently handles quantized KV cache.
    """
    if isinstance(cache, QuantizedKVCache):
        return quantized_scaled_dot_product_attention(queries, keys, values, cache=cache, scale=scale, mask=mask)
    else:
        return mx.fast.scaled_dot_product_attention(queries, keys, values, scale=scale, mask=mask)


import mlx.core as mx
import mlx.nn as nn

class LayerNorm2D(nn.Module):
    def __init__(self, d1, d2, eps):
        super().__init__()

        self.weight = mx.zeros((d1, d2))
        self.eps = eps

    def __call__(self, x):
        return self.weight * mx.fast.layer_norm(x, None, None, self.eps)
    
class RMSNorm(nn.Module):
    """
    Root Mean Square Normalization module.
    """
    def __init__(self,
                 dims: int,
                 eps: float = 1e-6):
        super().__init__()

        self.weight = mx.ones((dims,))
        self.eps = eps

    def __call__(self, x):
        return mx.fast.rms_norm(x, 1.0 + self.weight, self.eps)


import logging

import numpy as np

import mlx.core as mx
import mlx.nn as nn

from sillm.core.llm import LLM
from sillm.models.args import ModelArgs
from sillm.training.lora import TrainableLoRA

logger = logging.getLogger("sillm")

class DistillationLoRA(TrainableLoRA):
    """
    Trainable distillation LLM.
    References:
    https://arxiv.org/abs/1503.02531
    """
    @staticmethod
    def from_model(target_llm: LLM, draft_llm: LLM, **kwargs):
        """
        Convert LLM to trainable distillation LLM.
        Args:
            llm: LLM to convert.
        Returns:
            Trainable distillation LLM.
        """
        model = DistillationLoRA(target_llm.model, draft_llm.model, target_llm.tokenizer, draft_llm.tokenizer, target_llm.args, **kwargs)
        model._quantization = target_llm._quantization

        return model

    def __init__(self,
                 target_model,
                 draft_model,
                 target_tokenizer,
                 draft_tokenizer,
                 args: ModelArgs,
                 loss_alpha: float = 0.1
                 ):
        """
        Args:
            target_model: Target model instance.
            draft_model: Draft model instance.
            target_tokenizer: Target tokenizer instance.
            draft_tokenizer: Draft tokenizer instance.
            args: Model arguments.
            loss_alpha: Alpha parameter for distillation loss.
            loss_beta: Beta parameter for distillation loss.
        """
        if draft_tokenizer.vocab != target_tokenizer.vocab:
            raise ValueError("Target and draft tokenizers must have the same vocabulary")
        self.vocab_size = len(target_tokenizer.vocab)

        super().__init__(target_model, target_tokenizer, args)

        self.draft_model = draft_model
        self.draft_tokenizer = draft_tokenizer

        self.alpha = loss_alpha
        self.beta = 1.0 - loss_alpha

        logger.info(f"Initialized model distillation with alpha {loss_alpha}")
        
    def loss(self,
             inputs: mx.array,
             targets: mx.array,
             loss_masks: mx.array
             ):
        num_tokens = loss_masks.sum()

        # Forward pass for student model
        student_logits = self.model(inputs)[:, :, :self.vocab_size]
        student_probs = nn.softmax(student_logits)

        # Calculate student loss
        student_loss = nn.losses.cross_entropy(student_logits, targets) * loss_masks
        student_loss = student_loss.sum() / num_tokens

        # Forward pass for teacher model
        teacher_logits = self.draft_model(inputs)[:, :, :self.vocab_size]
        teacher_probs = nn.softmax(teacher_logits)

        # Calculate distillation loss
        distill_loss = nn.losses.cross_entropy(student_probs, teacher_probs) * loss_masks
        distill_loss = distill_loss.sum() / num_tokens

        # Combine losses
        loss_value = self.alpha * student_loss + self.beta * distill_loss
        reward = mx.stack([student_loss, distill_loss])
        
        return loss_value, reward, num_tokens


import argparse
import pathlib

import huggingface_hub

if __name__ == "__main__":
    # Parse commandline arguments
    parser = argparse.ArgumentParser(description="Downloader for Hugging Face models.")
    parser.add_argument("model", type=str, help="The model identifier (e.g. microsoft/Phi-3-medium-4k-instruct)")
    parser.add_argument("output", type=str, help="Output directory for model files")
    parser.add_argument("--token", default=None, type=str, help="Hugging Face API token")
    parser.add_argument("--proxy", default=None, type=str, help="Proxy URL")
    args = parser.parse_args()

    if args.token:
        huggingface_hub.login(args.token)

    model_path = pathlib.Path(args.output)
    model_path.mkdir(parents=True, exist_ok=True)

    allow_patterns = ["*.json", "*.safetensors", "*.model", "pytorch_model*.bin"]

    if args.proxy:
        proxies = {
            "https": args.proxy
        }
    else:
        proxies = None

    huggingface_hub.snapshot_download(repo_id=args.model, allow_patterns=allow_patterns, proxies=proxies, local_dir=model_path)


import math
from typing import Union, List

import mlx.core as mx
import mlx.nn as nn

from sillm.models.args import ModelArgs

def init_rope(args: ModelArgs):
    rope_type = "default"
    rope_scale = 1.0
    if args.rope_scaling is not None:
        rope_type = (args.rope_scaling.get("type") or args.rope_scaling.get("rope_type") or "default")

        if rope_type == "linear":
            rope_scale = 1 / args.rope_scaling["factor"]

    rope_dims = args.head_dim
    if args.partial_rotary_factor is not None:
        rope_dims = int(rope_dims * args.partial_rotary_factor)

    if rope_type in ("default", "linear"):
        return nn.RoPE(rope_dims,
                       traditional=args.rope_traditional,
                       base=args.rope_theta,
                       scale=rope_scale)
    elif rope_type == "llama3":
        return Llama3RoPE(rope_dims,
                          max_position_embeddings=args.max_position_embeddings,
                          traditional=args.rope_traditional,
                          base=args.rope_theta,
                          scale=rope_scale,
                          rope_scaling=args.rope_scaling)
    elif rope_type in ("su", "longrope"):
        return SuScaledRotaryEmbedding(rope_dims,
                                       max_position_embeddings=args.max_position_embeddings,
                                       original_max_position_embeddings=args.original_max_position_embeddings,
                                       base=args.rope_theta,
                                       short_factor=args.rope_scaling["short_factor"],
                                       long_factor=args.rope_scaling["long_factor"],
                                       short_mscale=args.rope_scaling.get("short_mscale", None),
                                       long_mscale=args.rope_scaling.get("long_mscale", None)
                                       )
    elif args.rope_scaling["type"] == "yarn":
        raise NotImplementedError("Yarn RoPE is not implemented")
    else:
        raise NotImplementedError(f"Unknown scaling type {rope_scale}")

########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/85dc76f6e0f2cf3ee3d84c211868a6856e163f3f/llms/mlx_lm/models/llama.py#L49
########
class Llama3RoPE(nn.Module):
    def __init__(self,
                 head_dim: int,
                 max_position_embeddings: int = 131072,
                 traditional: bool = True,
                 base: float = 10000,
                 scale: float = 1.0,
                 rope_scaling: dict = None,
                 ):
        super().__init__()

        self.head_dim = head_dim
        self.max_position_embeddings = max_position_embeddings
        self.traditional = traditional
        self.original_base = base
        self.scale = scale

        ########
        # Calculate base frequencies
        # References:
        # https://github.com/huggingface/transformers/blob/d5a99dfcee6e94065cb7c83cc8ab6fc5daa0cc4e/src/transformers/modeling_rope_utils.py#L318
        ########
        factor = rope_scaling.get("factor", 8.0)
        low_freq_factor = rope_scaling.get("low_freq_factor", 1.0)
        high_freq_factor = rope_scaling.get("high_freq_factor", 4.0)
        old_context_len = rope_scaling.get("original_max_position_embeddings", 8192)

        low_freq_wavelen = old_context_len / low_freq_factor
        high_freq_wavelen = old_context_len / high_freq_factor

        freqs = self.original_base ** (mx.arange(0, head_dim, 2) / head_dim)
        wavelens = 2 * mx.pi * freqs
        new_base_freqs = []

        smooths = (wavelens - high_freq_wavelen) / (low_freq_wavelen - high_freq_wavelen)
        new_base_freqs = freqs * (1 - smooths) * factor + smooths
        new_base_freqs = mx.where(wavelens < high_freq_wavelen, freqs, new_base_freqs)
        new_base_freqs = mx.where(wavelens > low_freq_wavelen, freqs * factor, new_base_freqs)

        self.base = new_base_freqs.mean().item()
            
    def __call__(self, x, offset: int = 0):
        seq_len = x.shape[1] + offset
        base = self.base
        if self.max_position_embeddings and seq_len > self.max_position_embeddings:
            base *= (
                (self.scale * seq_len / self.max_position_embeddings) - (self.scale - 1)
            ) ** (self.head_dim / (self.head_dim - 2))

        return mx.fast.rope(
            x,
            self.head_dim,
            traditional=self.traditional,
            base=base,
            scale=self.scale,
            offset=offset,
        )

########
# Su Scaled Rotary Embedding
# References:
# https://huggingface.co/microsoft/Phi-3-mini-128k-instruct/blob/38143357bf52ce57009ecbd58cf9f0b0029cb393/modeling_phi3.py#L142
# https://github.com/ml-explore/mlx-examples/blob/3c6e8b11af9dda55ee8d38ee9f612a65118f7793/llms/mlx_lm/models/su_rope.py#L10
########
class SuScaledRotaryEmbedding(nn.Module):
    """
    Scaled Uniform RoPE.
    """
    def __init__(self,
                 head_dim: int,
                 max_position_embeddings: int = 131072,
                 original_max_position_embeddings: int = 4096,
                 base: float = 10000,
                 scale: float = 1.0,
                 short_factor: Union[List[float], float] = 1.0,
                 long_factor: Union[List[float], float] = 1.0,
                 short_mscale: float = None,
                 long_mscale: float = None,
                 ):
        super().__init__()

        self.original_max_position_embeddings = original_max_position_embeddings

        self._inv_freq_short = 1.0 / (
            mx.array(short_factor, dtype=mx.float32)
            * base ** (mx.arange(0, head_dim, 2, dtype=mx.float32) / head_dim)
        )
        self._inv_freq_long = 1.0 / (
            scale
            * mx.array(long_factor, dtype=mx.float32)
            * base ** (mx.arange(0, head_dim, 2, dtype=mx.float32) / head_dim)
        )
        self.scaling_factor = long_mscale or math.sqrt(1 + math.log(max_position_embeddings / self.original_max_position_embeddings) / math.log(self.original_max_position_embeddings))

    def _get_cos_sin(self, offset, L):
        position_ids = mx.arange(offset, offset + L, dtype=mx.float32)
        inv_freq = (
            self._inv_freq_long
            if (offset + L) > self.original_max_position_embeddings
            else self._inv_freq_short
        )
        freqs = position_ids[:, None] * inv_freq[None, :]
        emb = mx.concatenate([freqs, freqs], axis=-1)
        cos = mx.cos(emb) * self.scaling_factor
        sin = mx.sin(emb) * self.scaling_factor
        return cos, sin

    def __call__(self, x, offset: int = 0):
        def _rotate_half(_x):
            midpoint = _x.shape[-1] // 2
            x1, x2 = _x[..., :midpoint], _x[..., midpoint:]
            return mx.concatenate([-x2, x1], axis=-1)

        cos, sin = self._get_cos_sin(offset, x.shape[2])
        return (x * cos) + (_rotate_half(x) * sin)


from typing import Optional

import mlx.core as mx
import mlx.nn as nn

from sillm.core.cache import KVCache
from sillm.models.base import BaseModel
from sillm.models.args import ModelArgs
from sillm.modules.act import init_act
import sillm.models.llama as llama

class FeedForward(nn.Module):
    """
    Feed-forward module.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        super().__init__()

        self.w2 = nn.Linear(args.hidden_dim, args.dim, bias=True)
        self.w3 = nn.Linear(args.dim, args.hidden_dim, bias=True)
        
        self.act = init_act(args)
    
    def __call__(self,
                 x: mx.array
                 ) -> mx.array:
        """
        Args:
            x: Input tensor.
        Returns:
            Output tensor.
        """
        return self.w2(self.act(self.w3(x)))
    
class TransformerBlock(nn.Module):
    """
    Transformer block.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        super().__init__()
        self.args = args
        
        self.n_heads = args.n_heads
        self.dim = args.dim
        
        self.attention = llama.Attention(args=args)
        self.feed_forward = FeedForward(args=args)
        self.attention_norm = nn.LayerNorm(args.dim, eps=args.norm_eps, bias=args.mlp_bias)
        self.ffn_norm = nn.LayerNorm(args.dim, eps=args.norm_eps, bias=args.mlp_bias)

    def forward(
            self,
            x: mx.array,
            mask: Optional[mx.array] = None,
            cache: Optional[KVCache] = None,
            ) -> mx.array:
        """
        Args:
            x: Input tensor.
            mask: Mask tensor.
            cache: Cache from previous forward pass.
        Returns:
            Output tensor and cache.
        """
        r = self.attention(self.attention_norm(x), mask, cache)
        h = x + r
        r = self.feed_forward(self.ffn_norm(h))
        out = h + r
        
        return out

class Model(llama.Model):
    """
    Pharia-1 model wrapper.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        # Fix Pharia-1 model arguments
        args.rope_traditional = True

        BaseModel.__init__(self, args)
        self.args = args

        self.n_layers = args.n_layers
        self.vocab_size = args.vocab_size
        
        self.tok_embeddings = nn.Embedding(args.vocab_size, args.dim)
        self.layers = [TransformerBlock(args=args) for _ in range(args.n_layers)]
        self.norm = nn.LayerNorm(args.dim, eps=args.norm_eps)
        self.output = nn.Linear(args.dim, args.vocab_size, bias=False)


import time
import logging

import mlx.core as mx

import sillm
from sillm.core.llm import LLM
from sillm.models.args import ModelArgs

logger = logging.getLogger("sillm")

class SpeculativeLLM(LLM):
    @staticmethod
    def from_models(draft_llm: LLM, target_llm: LLM):
        return SpeculativeLLM(draft_llm, target_llm)

    def __init__(self, draft_llm: LLM, target_llm: LLM):
        self.draft_model = draft_llm.model
        self.target_model = target_llm.model
        self.tokenizer = draft_llm.tokenizer
        self.args = draft_llm.args

    def generate(self,
                 prompt: str,
                 lookahead: int = 8,
                 temperature: float = 0.0,
                 max_tokens: int = 2048,
                 token_ids: bool = False,
                 ):
        start = time.perf_counter()
        
        # Tokenize prompt
        input_tokens = self.tokenizer.encode(prompt)
            
        # Initialize metadata
        timing = {
            "runtime": 0.0,
            "eval_time": 0.0,
            "tokenizer_time": time.perf_counter() - start
        }
        usage = {
            "prompt_tokens": len(input_tokens),
            "completion_tokens": 0,
            "total_tokens": 0
        }
        speculative = {
            "num_accepted": 0,
            "drafted": []
        }
        metadata = {
            "timing": timing,
            "usage": usage,
            "speculative": speculative,
            "token_ids": []
        }

        # Define stop tokens
        stop_tokens = self.tokenizer.special_ids

        def sample(logits):
            if temperature == 0:
                return mx.argmax(logits, axis=-1)
            else:
                return mx.random.categorical(logits * (1 / temperature))

        # Initialize KV caches
        draft_cache = sillm.KVCache.for_model(self.draft_model)
        target_cache = sillm.KVCache.for_model(self.target_model)

        # Initial forward pass through draft and target models
        inputs = mx.array(input_tokens[:-1])[None]
        draft_logits = self.draft_model(inputs, cache=draft_cache)
        target_logits = self.target_model(inputs, cache=target_cache)

        # Calculate evaluation time
        mx.eval(draft_logits, target_logits)
        timing["eval_time"] = time.perf_counter() - start

        # Main generation loop
        tokens = input_tokens
        text = self.tokenizer.decode(tokens)
        while True:
            # Draft speculative tokens
            y = mx.array(tokens[-1:])
            draft_tokens = []
            for _ in range(lookahead):
                draft_logits = self.draft_model(y[None], cache=draft_cache)
                draft_logits = draft_logits[:, -1, :]

                y = sample(draft_logits)
                draft_tokens.append(y.item())

                if y.item() in stop_tokens:
                    break
            draft_indices = mx.array(draft_tokens)
            
            # Check drafted tokens with the target model
            inputs = mx.array(tokens[-1:] + draft_tokens[:-1])[None]
            target_logits = self.target_model(inputs, cache=target_cache)

            # Accept or reject draft tokens
            target_indices = sample(target_logits).flatten()
            target_tokens = target_indices.tolist()
            accept_mask = (target_indices == draft_indices).tolist()
            num_accepted = (accept_mask + [False]).index(False)

            if num_accepted == 0:
                # All draft tokens were rejected - sample one token from target model
                result_tokens = [target_tokens[0]]
                speculative["drafted"] += [False]
            elif num_accepted < len(draft_tokens):
                # Some draft tokens were accepted - sample last token from target model
                result_tokens = draft_tokens[:num_accepted] + [target_tokens[num_accepted]]
                speculative["drafted"] += [True] * num_accepted + [False]
            else:
                # All draft tokens were accepted
                result_tokens = draft_tokens
                speculative["drafted"] += [True] * num_accepted

            # Check for completion conditions
            for i, t in enumerate(result_tokens):
                if len(tokens) + i >= max_tokens:
                    metadata["finish_reason"] = "length"
                    result_tokens = result_tokens[:i]
                if t in stop_tokens:
                    metadata["finish_reason"] = "stop"
                    result_tokens = result_tokens[:i]

            # Update tokens
            tokens += result_tokens

            # Update metadata
            usage["completion_tokens"] += len(result_tokens)
            usage["total_tokens"] = len(tokens)
            timing["runtime"] = time.perf_counter() - start
            speculative["num_accepted"] += num_accepted
            if token_ids:
                metadata["token_ids"] = tokens

            # Adjust cache offsets
            for dc, tc in zip(draft_cache, target_cache):
                dc.offset = len(tokens) - 1
                tc.offset = len(tokens) - 1

            # Update text
            text_offset = len(text)
            text = self.tokenizer.decode(tokens)

            # Yield result
            yield text[text_offset:], metadata

            # Check for completion
            if "finish_reason" in metadata:
                break

class SpeculativeEdit(LLM):
    @staticmethod
    def from_model(llm: LLM):
        return SpeculativeEdit(llm)

    def __init__(self, model: LLM):
        self.target_model = model
        self.model = model.model
        self.tokenizer = model.tokenizer
        self.args = model.args

        self._quantization = model._quantization
        self._mapping = model._mapping

    def generate(self,
                 prompt: str,
                 target: str,
                 lookahead: int = 32,
                 key_size: int = 3,
                 temperature: float = 0.0,
                 max_tokens: int = 8192,
                 token_ids: bool = False,
                 ):
        start = time.perf_counter()
        
        # Tokenize prompt
        input_tokens = self.tokenizer.encode(prompt)

        # Initialize metadata
        timing = {
            "runtime": 0.0,
            "eval_time": 0.0,
            "tokenizer_time": time.perf_counter() - start
        }
        usage = {
            "prompt_tokens": len(input_tokens),
            "completion_tokens": 0,
            "total_tokens": 0
        }
        speculative = {
            "num_drafted": 0,
            "num_accepted": 0,
            "drafted": []
        }
        metadata = {
            "timing": timing,
            "usage": usage,
            "speculative": speculative,
            "token_ids": []
        }

        # Handle non-speculative generation
        if lookahead <= 1:
            for s, metadata in self.target_model.generate(prompt, temperature=temperature, max_tokens=max_tokens, token_ids=token_ids):
                metadata["speculative"] = speculative
                yield s, metadata

            return

        # Define stop tokens
        stop_tokens = self.tokenizer.special_ids

        def sample(logits):
            if temperature == 0:
                return mx.argmax(logits, axis=-1)
            else:
                return mx.random.categorical(logits * (1 / temperature))

        # Initialize KV caches
        cache = sillm.KVCache.for_model(self.model)

        # Initial forward pass
        inputs = mx.array(input_tokens[:-1])[None]
        logits = self.model(inputs, cache=cache)

        # Calculate evaluation time
        mx.eval(logits)
        timing["eval_time"] = time.perf_counter() - start

        # Initialize speculative mode
        target_tokens = input_tokens[-1:] + self.tokenizer.encode(target, bos=False, eos=True) + [self.tokenizer.eos_id] * lookahead
        spec_offset = 0
        spec_mode = True
        total_drafted = 0
        total_accepted = 0

        # Main generation loop
        tokens = input_tokens
        text = self.tokenizer.decode(tokens)
        while True:
            if spec_mode and lookahead > 1:
                # Speculate based on target text
                inputs = mx.array(target_tokens[spec_offset:spec_offset+lookahead])[None]
                logits = self.model(inputs, cache=cache)

                spec_tokens = mx.array(target_tokens[spec_offset+1:spec_offset+lookahead+1])
                edit_tokens = mx.argmax(logits, axis=-1).flatten()[:len(spec_tokens)]
                accept_mask = (spec_tokens == edit_tokens).tolist()
                num_accepted = (accept_mask + [False]).index(False)

                if num_accepted == 0:
                    # All target tokens were rejected - sample first token from the model
                    logits = logits[:, 0, :]
                    y = sample(logits)

                    result_tokens = [y.item()]
                    speculative["drafted"] += [False]
                elif num_accepted < lookahead:
                    # Some target tokens were accepted - sample last token from the model
                    result_tokens = spec_tokens[:num_accepted].tolist() + [edit_tokens[num_accepted].item()]
                    speculative["drafted"] += [True] * num_accepted + [False]
                else:
                    # All target tokens were accepted
                    result_tokens = spec_tokens[:num_accepted].tolist()
                    speculative["drafted"] += [True] * num_accepted

                # Update speculative mode
                spec_mode = (num_accepted == lookahead)
                spec_offset += num_accepted

                # Update draft counter
                total_drafted += lookahead
            else:
                # Generate next token
                inputs = mx.array(tokens[-1:])[None]
                logits = self.model(inputs, cache=cache)
                y = sample(logits)

                result_tokens = [y.item()]
                speculative["drafted"] += [False]
                num_accepted = 0

                # Try to recover speculative mode
                spec_key = tokens[-key_size:]
                for i in range(total_accepted, len(target_tokens) - lookahead - key_size):
                    if target_tokens[i:i+key_size] == spec_key:
                        spec_offset = i + key_size
                        spec_mode = True

            # Check for completion conditions
            for i, t in enumerate(result_tokens):
                if len(tokens) + i >= max_tokens:
                    metadata["finish_reason"] = "length"
                    result_tokens = result_tokens[:i]
                if t in stop_tokens:
                    metadata["finish_reason"] = "stop"
                    result_tokens = result_tokens[:i]

            # Update tokens
            tokens += result_tokens
            total_accepted += num_accepted
            
            # Update metadata
            usage["completion_tokens"] += len(result_tokens)
            usage["total_tokens"] = len(tokens)
            timing["runtime"] = time.perf_counter() - start
            speculative["num_drafted"] = total_drafted
            speculative["num_accepted"] = total_accepted
            if token_ids:
                metadata["token_ids"] = tokens

            # Adjust cache offsets
            for c in cache:
                c.offset = len(tokens) - 1

            # Update text
            text_offset = len(text)
            text = self.tokenizer.decode(tokens)

            # Yield result
            yield text[text_offset:], metadata

            # Check for completion
            if "finish_reason" in metadata:
                break


import argparse

import sillm
import sillm.utils as utils

if __name__ == "__main__":
    # Parse commandline arguments
    parser = argparse.ArgumentParser(description="Convert a model to a different format.")
    parser.add_argument("input", type=str, help="The input model directory or file")
    parser.add_argument("output", type=str, help="The output model directory or file")
    parser.add_argument("-a", "--input_adapters", default=None, type=str, help="Load LoRA adapter weights from .safetensors file")
    parser.add_argument("-r", "--remap", default=False, action="store_true", help="Remap weights keys to native SiLLM format")
    parser.add_argument("-q2", default=False, action="store_true", help="Quantize the model to 2 bits")
    parser.add_argument("-q3", default=False, action="store_true", help="Quantize the model to 3 bits")
    parser.add_argument("-q4", default=False, action="store_true", help="Quantize the model to 4 bits")
    parser.add_argument("-q6", default=False, action="store_true", help="Quantize the model to 6 bits")
    parser.add_argument("-q8", default=False, action="store_true", help="Quantize the model to 8 bits")
    parser.add_argument("-dtype", type=str, default=None, help="Cast model weights to specified data type")
    parser.add_argument("-m", "--max_shard_size", type=int, default=5368709120, help="Maximum shard size in bytes")
    parser.add_argument("-v", "--verbose", default=1, action="count", help="Increase output verbosity")
    args = parser.parse_args()

    # Initialize logging
    log_level = 40 - (10 * args.verbose) if args.verbose > 0 else 0
    logger = utils.init_logger(log_level)

    # Log commandline arguments
    if log_level <= 10:
        utils.log_arguments(args.__dict__)

    # Load model
    model = sillm.load(args.input)

    if args.input_adapters is not None:
        # Convert model to trainable
        model = sillm.TrainableLoRA.from_model(model)

        lora_config = model.load_lora_config(args.input_adapters)

        # Initialize LoRA layers
        model.init_lora(**lora_config)

        # Load and merge adapter file
        model.load_adapters(args.input_adapters)
        model.merge_and_unload_lora()

    # Quantize model or cast model weights
    if args.q2 is True:
        model.quantize(bits=2)
    elif args.q3 is True:
        model.quantize(bits=3)
    elif args.q4 is True:
        model.quantize(bits=4)
    elif args.q6 is True:
        model.quantize(bits=6)
    elif args.q8 is True:
        model.quantize(bits=8)
    elif args.dtype is not None:
        model.astype(args.dtype)

    # Disable mapping to old keys
    if args.remap:
        model._mapping = None

    # Save model
    model.save(args.output, max_shard_size=args.max_shard_size)


from functools import partial

import mlx.core as mx
import mlx.nn as nn

@partial(mx.compile, inputs=mx.random.state, outputs=mx.random.state)
def top_k(logits: mx.array,
          k: int = 1
          ) -> mx.array:
    top_k_indices = mx.argpartition(logits, k, axis=-1)[:, -k:]
    mask = mx.zeros_like(logits)
    mask[:, top_k_indices] = 1.0

    return logits * mask

@partial(mx.compile, inputs=mx.random.state, outputs=mx.random.state)
def top_p(logits: mx.array,
          p: float = 0.9
          ) -> mx.array:
    probs = mx.softmax(logits, axis=-1)
    sorted_indices = mx.argsort(probs, axis=-1).squeeze(0)
    sorted_probs = probs[..., sorted_indices]
    cum_probs = mx.cumsum(sorted_probs, axis=-1)
    
    mask = cum_probs <= p
    logits = mx.where(mask[..., sorted_indices], logits, 0.0)

    return logits

########
# References:
# Chenxia Tang, Jianchun Liu, Hongli Xu, Liusheng Huang. Top-nσ: Not All Logits Are You Need. https://arxiv.org/abs/2411.07641
########
@partial(mx.compile, inputs=mx.random.state, outputs=mx.random.state)
def top_nsigma(logits: mx.array,
               n: float = 1.0
               ) -> mx.array:
    threshold = logits.max(axis=-1, keepdims=True) - n * logits.std(axis=-1)
    mask = logits >= threshold
    logits = mx.where(mask, logits, 0.0)

    return logits

@partial(mx.compile, inputs=mx.random.state, outputs=mx.random.state)
def apply_repetition_penalty(logits: mx.array,
                             tokens: list,
                             repetition_penalty: float = 1.0,
                             repetition_window: int = 25
                             ) -> mx.array:
    indices = mx.array(tokens[-repetition_window:])
    repeated_logits = logits[:, indices]
    repeated_logits = mx.where(repeated_logits < 0, repeated_logits * repetition_penalty, repeated_logits / repetition_penalty)
    logits[:, indices] = repeated_logits

    return logits


import logging
import math

import mlx.core as mx
import mlx.nn as nn
from mlx.utils import tree_flatten

logger = logging.getLogger("sillm")

class SparseAutoencoder(nn.Module):
    """
    Sparse Autoencoder
    References:
    https://arxiv.org/pdf/2309.08600
    https://arxiv.org/pdf/2406.04093v1
    """
    def __init__(self,
                 dim: int,
                 hidden_dim: int
                 ):
        super().__init__()

        self.num_features = hidden_dim
        logger.info(f"Initialized Sparse Autoencoder with input dimensions {dim} and {hidden_dim} features")

    def encode(self, x: mx.array):
        raise NotImplementedError("Method encode is not implemented")
    
    def decode(self, x: mx.array):
        raise NotImplementedError("Method decode is not implemented")
    
    def save_weights(self,
                     weights_path: str
                     ):
        """
        Save model weights into a single safetensors file.
        Args:
            weights_path: Path to weights file.
        """
        state = dict(tree_flatten(self.model.parameters()))
        mx.save_safetensors(weights_path, state)

        logger.info(f"Saved weights for Sparse Autoencoder to {weights_path}")

    @staticmethod
    def load(model_path: str):
        """
        Initialize sparse autoencoder from weights file
        Args:
            args: LLM config arguments
            model_path: Path to sparse autoencoder weights
        """
        weights = mx.load(model_path)

        if "threshold" in weights:
            hidden_dim, dim = weights["W_enc"].shape
            
            model = JumpReLUSparseAutoencoder(dim, hidden_dim)
        elif "encoder.weight" in weights:
            hidden_dim, dim = weights["encoder.weight"].shape

            model = EleutherSparseAutoencoder(dim, hidden_dim)
        else:
            raise ValueError("Unknown type of Sparse Autoencoder")
        
        model.update(weights)
        logger.info(f"Loaded weights for Sparse Autoencoder from {model_path}")

        return model

class JumpReLUSparseAutoencoder(SparseAutoencoder):
    """
    JumpReLU Sparse Autoencoder
    References:
    https://storage.googleapis.com/gemma-scope/gemma-scope-report.pdf
    https://colab.research.google.com/drive/17dQFYUYnuKnP6OwQPH9v_GSYUW5aj-Rp
    """
    def __init__(self,
                 dim: int,
                 hidden_dim: int
                 ):
        super().__init__(dim, hidden_dim)

        scale = math.sqrt(1.0 / hidden_dim)
        self.W_dec = mx.random.uniform(-scale, scale, (hidden_dim, dim))
        self.b_dec = mx.zeros((dim,))

        self.W_enc = self.W_dec.T
        self.b_enc = mx.zeros((hidden_dim,))

        self.threshold = mx.full((hidden_dim), 0.001)

    def encode(self, x: mx.array):
        x = x @ self.W_enc + self.b_enc
        mask = (x > self.threshold)

        return mask * nn.relu(x)
    
    def decode(self, x: mx.array):
        return x @ self.W_dec + self.b_dec

class EleutherSparseAutoencoder(SparseAutoencoder):
    """
    Eleuther Sparse Autoencoder
    References:
    https://github.com/EleutherAI/sae
    """
    def __init__(self,
                 dim: int,
                 hidden_dim: int
                 ):
        super().__init__(dim, hidden_dim)

        self.encoder = nn.Linear(dim, hidden_dim)

        scale = math.sqrt(1.0 / hidden_dim)
        self.W_dec = mx.random.uniform(-scale, scale, (hidden_dim, dim))
        self.b_dec = mx.zeros((dim,))

    def encode(self, x: mx.array):
        x = x - self.b_dec
        x = self.encoder(x)
        
        return nn.relu(x)


import os
import logging

import mlx.core as mx

logger = logging.getLogger("sillm")

def set_memory_limit(memory_limit: float = 0.9,
                     relaxed: bool = False
                     ):
    """
    Set memory limit.
    Args:
        memory_limit: The memory limit as percentage of total system memory.
    """
    # Get system memory
    system_memory = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")

    logger.info(f"Setting memory limit to {memory_limit:.0%} of system memory")
    mx.metal.set_memory_limit(int(memory_limit * system_memory), relaxed=relaxed)


import math

import mlx.core as mx
import mlx.nn as nn

from .act import init_act

########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/f530f56df2738a54982c4541189a8c8d7cd94c44/llms/mlx_lm/models/switch_layers.py#L9
########
class QuantizedSwitchLinear(nn.Module):
    def __init__(self,
                 input_dims: int,
                 output_dims: int,
                 num_experts: int,
                 bias: bool = True,
                 group_size: int = 64,
                 bits: int = 4,
                 ):
        super().__init__()

        scale = math.sqrt(1 / input_dims)
        self.weight, self.scales, self.biases = mx.quantize(
            mx.random.uniform(
                low=-scale,
                high=scale,
                shape=(num_experts, output_dims, input_dims),
            ),
            group_size=group_size,
            bits=bits,
        )

        if bias:
            self.bias = mx.zeros((num_experts, output_dims))

        self.group_size = group_size
        self.bits = bits

        # Freeze this model's parameters
        self.freeze()

    def unfreeze(self, *args, **kwargs):
        super().unfreeze(*args, **kwargs)
        self.freeze(recurse=False)

    @property
    def input_dims(self):
        return self.scales.shape[2] * self.group_size

    @property
    def output_dims(self):
        return self.weight.shape[1]

    @property
    def num_experts(self):
        return self.weight.shape[0]

    def __call__(self, x, indices):
        x = mx.gather_qmm(x, self["weight"], self["scales"], self["biases"], rhs_indices=indices, transpose=True, group_size=self.group_size, bits=self.bits)
        if "bias" in self:
            x = x + mx.expand_dims(self["bias"][indices], -2)
        return x

########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/f530f56df2738a54982c4541189a8c8d7cd94c44/llms/mlx_lm/models/switch_layers.py#L75
########
class SwitchLinear(nn.Module):
    def __init__(self,
                 input_dims: int,
                 output_dims: int,
                 num_experts: int,
                 bias: bool = True
                 ):
        super().__init__()

        scale = math.sqrt(1 / input_dims)
        self.weight = mx.random.uniform(low=-scale, high=scale, shape=(num_experts, output_dims, input_dims))

        if bias:
            self.bias = mx.zeros((num_experts, output_dims))

    @property
    def input_dims(self):
        return self.weight.shape[2]

    @property
    def output_dims(self):
        return self.weight.shape[1]

    @property
    def num_experts(self):
        return self.weight.shape[0]

    def __call__(self, x, indices):
        x = mx.gather_mm(x, self["weight"].swapaxes(-1, -2), rhs_indices=indices)
        if "bias" in self:
            x = x + mx.expand_dims(self["bias"][indices], -2)

        return x
    
    def to_quantized(self,
                     group_size: int = 64,
                     bits: int = 4
                     ) -> QuantizedSwitchLinear:
        num_experts, output_dims, input_dims = self.weight.shape
        
        ql = QuantizedSwitchLinear(input_dims, output_dims, num_experts, False, group_size, bits)
        ql.weight, ql.scales, ql.biases = mx.quantize(self.weight, group_size, bits)
        if "bias" in self:
            ql.bias = self.bias

        return ql

########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/f530f56df2738a54982c4541189a8c8d7cd94c44/llms/mlx_lm/models/switch_layers.py#L119
########
class SwitchGLU(nn.Module):
    def __init__(self,
                 dim: int,
                 hidden_dim: int,
                 num_experts: int,
                 bias: bool = False,
                 activation = nn.silu,
                 ):
        super().__init__()

        self.w1 = SwitchLinear(dim, hidden_dim, num_experts, bias=bias)
        self.w2 = SwitchLinear(hidden_dim, dim, num_experts, bias=bias)
        self.w3 = SwitchLinear(dim, hidden_dim, num_experts, bias=bias)

        self.act = activation

    def __call__(self, x, indices) -> mx.array:
        x = mx.expand_dims(x, (-2, -3))

        x_up = self.w3(x, indices)
        x_gate = self.w1(x, indices)
        x = self.w2(self.act(x_gate) * x_up, indices)

        return x.squeeze(-2)


import time

import mlx.core as mx
import mlx.nn as nn

from sillm.core.llm import LLM
from sillm.core.cache import KVCache

def beam_search(model: LLM,
                prompt: str,
                beam_width: int = 4,
                min_choices: int = 1,
                length_penalty: float = 0.6,
                max_tokens: int = 8192,
                ):
    tokenizer = model.tokenizer
    
    start = time.perf_counter()

    # Tokenize prompt
    input_tokens = tokenizer.encode(prompt)
    
    # Initialize metadata
    timing = {
        "runtime": 0.0,
        "eval_time": 0.0,
        "tokenizer_time": time.perf_counter() - start
    }
    usage = {
        "prompt_tokens": len(input_tokens),
        "completion_tokens": 0,
        "total_tokens": 0
    }
    metadata = {
        "timing": timing,
        "usage": usage,
        "logprobs": [],
        "token_ids": [],
        "finish_reason": "length"
    }

    # Define stop tokens
    stop_tokens = tokenizer.special_ids

    # Adjust max tokens
    max_tokens = max_tokens - len(input_tokens)

    # Initialize KV caches
    cache = KVCache.for_model(model)

    # Initial forward pass
    inputs = mx.array(input_tokens)[None]
    logits = model.model(inputs, cache=cache)

    timing["eval_time"] = time.perf_counter() - start

    # First round of beam updates
    logits = logits[:, -1, :]
    logits = nn.log_softmax(logits, axis=-1)
    top_k_indices = mx.argpartition(logits, beam_width, axis=-1)[:, -beam_width:]
    beam_logprobs = mx.take_along_axis(logits, top_k_indices, axis=-1).squeeze(0).reshape(beam_width, 1, 1)

    beam_tokens = [[] for _ in range(beam_width)]
    for i in range(beam_width):
        token_id = top_k_indices[0, i].item()
        beam_tokens[i].append(token_id)

    # Initialize cache for beam width by repeating keys and values
    for layer_cache in cache:
        layer_cache.keys = mx.repeat(layer_cache.keys, repeats=beam_width, axis=0)
        layer_cache.values = mx.repeat(layer_cache.values, repeats=beam_width, axis=0)

    finished_beams = []
    finished_scores = []
    while True:
        inputs = mx.array(beam_tokens)[:, -1:]
        logits = model.model(inputs, cache=cache)
        logits = nn.log_softmax(logits, axis=-1)

        top_k_indices = mx.argpartition(logits, beam_width, axis=-1)[:, :, -beam_width:]
        top_k_logprobs = mx.take_along_axis(logits, top_k_indices, axis=-1)
        
        combined_logprobs = top_k_logprobs + beam_logprobs
        flat_indices = top_k_indices.reshape(-1)
        flat_scores = combined_logprobs.reshape(-1)

        result_indices = mx.argpartition(flat_scores, beam_width, axis=-1)
        beam_indices = result_indices // beam_width
        considered_tokens = mx.take(flat_indices, result_indices).tolist()

        new_beam_tokens = []
        cache_indices = []
        for i in range(len(result_indices)-1, 0, -1):
            beam_index = beam_indices[i].item()
            token = considered_tokens[i]
            tokens = beam_tokens[beam_index] + [token]
            
            logprob = flat_scores[result_indices[i]]
            # Normalize logprob with length penalty
            if length_penalty > 0.0:
                logprob /= len(tokens) ** length_penalty
            
            if token in stop_tokens:
                # Add completion to finished beams
                finished_beams.append(tokens)
                finished_scores.append(logprob.item())
            else:
                # Update active beams
                beam_logprobs[len(new_beam_tokens)] = logprob
                new_beam_tokens.append(tokens)
                cache_indices.append(beam_index)

            # Found enough new beams
            if len(new_beam_tokens) >= beam_width:
                break
        beam_tokens = new_beam_tokens

        # Stop if not enough active beams
        if len(new_beam_tokens) < beam_width:
            break

        # Stop if max tokens reached
        if len(beam_tokens[0]) >= max_tokens:
            finished_beams.extend(beam_tokens)
            finished_scores.extend(beam_logprobs.flatten().tolist())
            break

        # Stop if best beam is worse than best finished beam
        if len(finished_scores) >= min_choices and max(beam_logprobs) < max(finished_scores):
            break

        # Update cache
        new_cache = KVCache.for_model(model.model)
        cache_indices = mx.array(cache_indices)
        for l in range(len(cache)):
            new_cache[l].offset = cache[l].offset
            new_cache[l].keys = mx.take(cache[l].keys, cache_indices, axis=0)
            new_cache[l].values = mx.take(cache[l].values, cache_indices, axis=0)
        cache = new_cache

    metadata["choices"] = []
    for i in mx.argsort(mx.array(finished_scores)).tolist()[::-1]:
        finish_reason = "length"
        if finished_beams[i][-1] in stop_tokens:
            finished_beams[i] = finished_beams[i][:-1]
            finish_reason = "stop"
        else:
            finish_reason = "length"

        choice = {
            "text": tokenizer.decode(finished_beams[i]),
            "score": finished_scores[i],
            "finish_reason": finish_reason
        }
        metadata["choices"].append(choice)

    result_index = mx.argmax(mx.array(finished_scores)).item()
    result_tokens = finished_beams[result_index]
    result = tokenizer.decode(result_tokens)

    timing["runtime"] = time.perf_counter() - start
    usage["completion_tokens"] = len(result_tokens)
    usage["total_tokens"] = usage["prompt_tokens"] + usage["completion_tokens"]

    return result, metadata


from typing import Optional

import mlx.core as mx
import mlx.nn as nn

from sillm.models.base import BaseModel
from sillm.core.cache import KVCache
from sillm.models.args import ModelArgs
from sillm.modules.rope import init_rope
import sillm.models.llama as llama

class Attention(llama.Attention):
    """
    Multi-head attention module.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        nn.Module.__init__(self)
        self.args = args

        self.n_heads: int = args.n_heads
        self.n_kv_heads: int = args.n_kv_heads

        self.scale = args.attention_multiplier

        self.wq = nn.Linear(args.dim, args.n_heads * args.head_dim, bias=args.attention_bias)
        self.wk = nn.Linear(args.dim, args.n_kv_heads * args.head_dim, bias=args.attention_bias)
        self.wv = nn.Linear(args.dim, args.n_kv_heads * args.head_dim, bias=args.attention_bias)
        self.wo = nn.Linear(args.n_heads * args.head_dim, args.dim, bias=args.attention_bias)

        self.rope = init_rope(args)

class TransformerBlock(nn.Module):
    """
    Transformer block.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        nn.Module.__init__(self)
        self.args = args

        self.n_heads = args.n_heads
        self.dim = args.dim

        self.residual_multiplier = args.residual_multiplier
        
        self.attention = Attention(args=args)
        self.feed_forward = llama.FeedForward(args=args)
        self.attention_norm = nn.RMSNorm(args.dim, eps=args.norm_eps)
        self.ffn_norm = nn.RMSNorm(args.dim, eps=args.norm_eps)

    def forward(
            self,
            x: mx.array,
            mask: Optional[mx.array] = None,
            cache: Optional[KVCache] = None,
            ) -> mx.array:
        """
        Args:
            x: Input tensor.
            mask: Mask tensor.
            cache: Cache from previous forward pass.
        Returns:
            Output tensor and cache.
        """
        h = x + self.attention(self.attention_norm(x), mask, cache) * self.residual_multiplier
        out = h + self.feed_forward(self.ffn_norm(h)) * self.residual_multiplier
        
        return out

########
# References:
# https://github.com/huggingface/transformers/blob/main/src/transformers/models/granite/modeling_granite.py
########
class Model(llama.Model):
    """
    Granite model wrapper.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        BaseModel.__init__(self, args)
        self.args = args

        self.n_layers = args.n_layers
        self.vocab_size = args.vocab_size

        self.embedding_multiplier = args.embedding_multiplier
        self.logits_scaling = args.logits_scaling
        
        self.tok_embeddings = nn.Embedding(args.vocab_size, args.dim)
        self.layers = [TransformerBlock(args=args) for _ in range(args.n_layers)]
        self.norm = nn.RMSNorm(args.dim, eps=args.norm_eps)

        if args.tie_word_embeddings:
            self.output = None
        else:
            self.output = nn.Linear(args.dim, args.vocab_size, bias=False)

    def __call__(self,
                 inputs: mx.array,
                 cache = None
                 ):
        """
        Args:
            inputs: Input tokens.
            cache: Cache from previous forward pass.
        Returns:
            Output logits.
        """
        h = self.tok_embeddings(inputs) * self.embedding_multiplier

        mask = None
        if h.shape[1] > 1:
            mask = nn.MultiHeadAttention.create_additive_causal_mask(h.shape[1]).astype(h.dtype)

        if cache is None:
            cache = [None] * len(self.layers)

        for e, layer in enumerate(self.layers):
            h = layer.forward(h, mask, cache[e])

        if self.output is None:
            logits = self.tok_embeddings.as_linear(self.norm(h))
        else:
            logits = self.output(self.norm(h))

        logits = logits / self.logits_scaling

        return logits


import pathlib
import json
import logging
import dataclasses

from typing import Union

from sillm.utils.mapping import map_config

logger = logging.getLogger("sillm")

@dataclasses.dataclass
class ModelArgs:
    """
    Model arguments.
    """
    model_type: str
    dim: int
    n_layers: int
    head_dim: int
    n_heads: int
    n_kv_heads: int
    norm_eps: float = 1e-5
    hidden_dim: int = None
    vocab_size: int = -1
    rope_theta: float = 10000.0
    rope_traditional: bool = True
    rope_scaling: dict = None
    partial_rotary_factor: float = None
    hidden_act: str = None
    max_position_embeddings: int = 0
    original_max_position_embeddings: int = 0
    tie_word_embeddings: bool = False
    attention_bias: bool = False
    mlp_bias: bool = False
    bos_token_id: int = None
    eos_token_id: Union[int, list] = None
    pad_token_id: int = None
    quantization: dict = None

    def __repr__(self):
        return json.dumps(dataclasses.asdict(self), indent=4)
    
    def log_config(self):
        for k, v in dataclasses.asdict(self).items():
            if isinstance(v, dict):
                for k2, v2 in v.items():
                    logger.debug(f"Config {k}.{k2}: {v2}")
            else:
                logger.debug(f"Config {k}: {v}")
    
    def fix_config(self, weights):
        """
        Fix config with shape information from weights.
        """
        if self.hidden_dim is None and "layers.0.feed_forward.w1.weight" in weights:
            self.hidden_dim = weights["layers.0.feed_forward.w1.weight"].shape[0]
        if self.vocab_size <= 0 and "output.weight" in weights:
            self.vocab_size = weights["output.weight"].shape[0]

    def save_config(self, config_path):
        """
        Save model config to JSON file.
        Args:
            config_path: Path to config file.
        """
        config = dataclasses.asdict(self)
        
        # Remove None values
        for k in list(config.keys()):
            if config[k] is None:
                del config[k]

        with open(config_path, "w") as f:
            f.write(json.dumps(config, indent=4))

    @staticmethod
    def load_config(config):
        ArgsClass = None

        args_map = {
            "llama": ModelArgs,
            "mistral": ModelArgs,
            "gemma": ModelArgs,
            "mixtral": MixtralArgs,
            "phi": ModelArgs,
            "qwen2": ModelArgs,
            "starcoder2": ModelArgs,
            "dbrx": DbrxArgs,
            "cohere": CohereArgs,
            "phi3": Phi3Args,
            "gemma2": Gemma2Args,
            "phimoe": PhiMoEArgs,
            "pharia-v1": ModelArgs,
            "granite": GraniteArgs,
            "mobilellm": ModelArgs,
            "exaone": ModelArgs,
        }

        if "model_type" in config:
            model_type = config["model_type"]

            if model_type in args_map:
                ArgsClass = args_map[model_type]
            else:
                logger.warning(f"Unknown model type {model_type} - falling back to `default` config")
                ArgsClass = ModelArgs
        if ArgsClass is None:
            ArgsClass = ModelArgs
            config["model_type"] = "llama"
            logger.warning(f"No model type specified - falling back to `default` config")

        fields = ModelArgs.__annotations__
        fields.update(ArgsClass.__annotations__)
        config = {k:v for k, v in config.items() if k in fields}

        return ArgsClass(**config)
    
    @staticmethod
    def load_file(config_path):
        """
        Load model config from JSON file.
        Args:
            config_path: Path to config file.
        Returns:
            ModelArgs instance.
        """
        assert pathlib.Path(config_path).exists(), config_path

        with open(config_path, "r") as f:
            config = json.loads(f.read())
        config = map_config(config)

        return ModelArgs.load_config(config)

@dataclasses.dataclass
class MixtralArgs(ModelArgs):
    """
    Mixtral model arguments.
    """
    rope_theta: float = 1000000.0
    router_aux_loss_coef: float = 0.001
    moe: dict = None

    def __post_init__(self):
        if self.moe is None:
            self.moe = {
                "num_experts": 8,
                "num_experts_per_tok": 2
            }

@dataclasses.dataclass
class DbrxArgs(ModelArgs):
    """
    DBRX model arguments.
    """
    clip_qkv: int = 8
    rope_theta: float = 500000.0
    router_aux_loss_coef: float = 0.05
    moe: dict = None

    def __post_init__(self):
        if self.moe is None:
            self.moe = {
                "num_experts": 16,
                "num_experts_per_tok": 4
            }

        if self.bos_token_id is None:
            self.bos_token_id = 100257
        if self.eos_token_id is None:
            self.eos_token_id = 100257

@dataclasses.dataclass
class CohereArgs(ModelArgs):
    """
    Cohere model arguments.
    """
    norm_bias: bool = False
    use_qk_norm: bool = False
    logit_scale: float = 0.0625

@dataclasses.dataclass
class Phi3Args(ModelArgs):
    """
    Phi-3 model arguments.
    """
    embd_pdrop: float = 0.0

@dataclasses.dataclass
class Gemma2Args(ModelArgs):
    """
    Gemma 2 model arguments.
    """
    attn_logit_softcapping: float = 50.0
    final_logit_softcapping: float = 30.0
    query_pre_attn_scalar: float = 144.0

@dataclasses.dataclass
class PhiMoEArgs(ModelArgs):
    """
    Phi-MoE model arguments.
    """
    num_local_experts: int = 16
    num_experts_per_tok: int = 2
    rms_norm_eps: float = 1e-5

@dataclasses.dataclass
class GraniteArgs(ModelArgs):
    """
    Granite model arguments.
    """
    embedding_multiplier: float = 1.0
    residual_multiplier: float = 1.0
    attention_multiplier: float = 1.0
    logits_scaling: float = 1.0


import string
import re

import mlx.core as mx

class LogitFilter:
    """
    Logit mask filter.
    """
    def __init__(self,
                 tokenizer,
                 output_size: int
                 ):
        self.tokenizer = tokenizer
        self.output_size = output_size

    def reset(self):
        pass

    def __call__(self,
                 logits: mx.array
                 ) -> mx.array:
        raise NotImplementedError("Class LogitFilter is used for inheritance only")

class ASCIIFilter(LogitFilter):
    """
    Static logit mask filtering out tokens with non-ASCII printable characters.
    """
    def __init__(self,
                 tokenizer,
                 output_size: int
                 ):
        mask = mx.zeros(output_size)
        for i, s in enumerate(tokenizer.vocab_strings):
            if all(c in string.printable for c in s.strip()):
                mask[i] = 1.0

        for i in tokenizer.special_ids:
            mask[i] = 1.0

        self.mask = mask

    def __call__(self,
                 logits: mx.array
                 ) -> mx.array:
        return logits * self.mask


import os
import time
import logging

from functools import partial

import tqdm
import numpy as np

import mlx.core as mx
import mlx.nn as nn
import mlx.optimizers as optim
from mlx.utils import tree_map

from sillm.core.llm import LLM
from sillm.training.dataset import Dataset

logger = logging.getLogger("sillm")

class TrainableLLM(LLM):
    """
    Trainable LoRA model wrapper.
    """
    @staticmethod
    def from_model(llm: LLM):
        """
        Convert LLM to trainable LLM.
        Args:
            llm: LLM to convert.
        Returns:
            Trainable LLM.
        """
        return TrainableLLM(llm.model, llm.tokenizer, llm.args)
    
    def __init__(self, llm: LLM):
        """
        Args:
            tokenizer: Tokenizer instance.
            args: Model arguments.
        """
        self.model = llm.model
        self.tokenizer = llm.tokenizer
        self.args = llm.args

        self._quantization = llm._quantization
        self._mapping = llm._mapping

    def loss(self, *args, **kwargs):
        """
        Default loss function from model.
        """
        return self.model.loss(*args, **kwargs)
    
    ########
    # Based on mlx-examples:
    # https://github.com/ml-explore/mlx-examples/blob/e74889d0fa0fb49d95bfdf6a1dcad907713eb50e/lora/lora.py#L198
    ########
    def evaluate(self,
                 dataset: Dataset,
                 batch_size: int,
                 num_batches: int
                 ):
        """
        Evaluate model on dataset.
        Args:
            dataset: Dataset to evaluate on.
            batch_size: Batch size.
            num_batches: Number of batches to evaluate.
        Returns:
            Average loss.
        """
        losses = []
        for _, batch in zip(
            range(num_batches),
            dataset.iterate_batches(batch_size),
        ):
            loss_value, _, _ = self.loss(*batch)
            losses.append(loss_value.item())

        return np.mean(losses)
    
    ########
    # Based on mlx-examples:
    # https://github.com/ml-explore/mlx-examples/blob/e74889d0fa0fb49d95bfdf6a1dcad907713eb50e/lora/lora.py#L212
    ########
    def train(self, 
              dataset_training: Dataset,
              dataset_validation: Dataset,
              dataset_test: Dataset = None,
              batch_size: int = 4,
              optimizer_type: str = "adam",
              learning_rate: float = 1e-5,
              learning_decay: float = 0.0,
              learning_warmup: int = 0,
              compiled_step: bool = True,
              gradient_checkpointing: bool = False,
              gradient_accumulation_steps: int = 1,
              gradient_max_norm: float = 1.0,
              epochs: int = 1,
              iterations: int = 0,
              report_steps: int = 10,
              report_callback: callable = None,
              eval_steps: int = 100,
              eval_callback: callable = None,
              validation_samples: int = 40
              ):
        """
        Train model.
        Args:
            dataset_training: Training dataset.
            dataset_validation: Validation dataset.
            batch_size: Batch size.
            learning_rate: Learning rate.
            epochs: Number of epochs.
            iterations: Number of iterations.
            report_steps: Report every `report_steps` iterations.
            eval_steps: Evaluate every `eval_steps` iterations.
            eval_callback: Callback after eval.
            validation_samples: Number of validation samples.
            debug: Whether to enable debug mode.
        """
        # Calculate number of iterations
        if iterations == 0:
            iterations = len(dataset_training) // batch_size
        
        # Calculate number of validation batches
        validation_batches = validation_samples // batch_size
        
        logger.info(f"Training the model for {epochs} epochs of {iterations} batch iterations with batch size {batch_size}")
        logger.debug(f"Training learning rate: {learning_rate}")

        # Get system memory
        system_memory = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")

        # Initialize scheduler
        if learning_decay > 0.0:
            scheduler = optim.step_decay(learning_rate, 1-learning_decay, 1)

            logger.debug(f"Training learning decay: {learning_decay}")
        else:
            scheduler = optim.linear_schedule(learning_rate, learning_rate, 1)
        if learning_warmup > 0:
            warmup = optim.linear_schedule(learning_rate * 0.1, learning_rate, learning_warmup)
            scheduler = optim.join_schedules([warmup, scheduler], [learning_warmup])

            logger.debug(f"Training learning warmup steps: {learning_warmup}")

        # Initialize optimizer
        optimizer_type = optimizer_type.lower()
        if optimizer_type == "adam":
            optimizer = optim.Adam(learning_rate=scheduler)
        elif optimizer_type == "adamw":
            optimizer = optim.AdamW(learning_rate=scheduler, weight_decay=0.0)
        elif optimizer_type == "adafactor":
            optimizer = optim.Adafactor(learning_rate=scheduler, weight_decay=0.0)
        else:
            raise ValueError(f"Unknown optimizer type: {optimizer_type}")
        
        # Initialize gradient accumulation
        if gradient_accumulation_steps > 1:
            accum_grad = None
            accum_scale = 1 / gradient_accumulation_steps

            if gradient_checkpointing:
                gradient_checkpointing = False
                logger.warning(f"Gradient accumulation requires disabling gradient checkpointing")
            if compiled_step:
                compiled_step = False
                logger.warning(f"Gradient accumulation requires disabling compiled step function")

            logger.info(f"Enabled gradient accumulation with {gradient_accumulation_steps} steps")

        # Initialize gradient checkpointing
        if gradient_checkpointing:
            if not compiled_step:
                logger.warning(f"Gradient checkpointing requires compiled step function")
                compiled_step = True
            
            for layer in self.model.layers:
                layer.forward = nn.utils.checkpoint(layer, layer.forward)

            logger.info(f"Enabled gradient checkpointing")

        # Create value and gradient function for loss
        loss_value_and_grad = nn.value_and_grad(self.model, self.loss)

        # Initialize compiled step function
        if compiled_step:
            state = [self.model.state, optimizer.state]

            # Step function for forward and backward pass
            @partial(mx.compile, inputs=state, outputs=state)
            def step(batch):
                (loss_value, reward, num_tokens), grad = loss_value_and_grad(*batch)
                optimizer.update(self.model, grad)

                return loss_value, reward, num_tokens

        # Initialize variables for training loop
        losses = []
        rewards = None
        intv_tokens = 0

        # Main training loop
        start = time.perf_counter()
        pbar_epochs = tqdm.tqdm(range(epochs), desc="Epoch")
        for epoch in pbar_epochs:
            pbar_iterations = tqdm.tqdm(range(iterations), desc="Iter.", leave=False)
            for iter in pbar_iterations:
                n = epoch * iterations + iter
                batch = next(dataset_training.iterate_batches(batch_size, train=True))

                if compiled_step:
                    loss_value, reward, num_tokens = step(batch)
                else:
                    (loss_value, reward, num_tokens), grad = loss_value_and_grad(*batch)

                    if gradient_accumulation_steps > 1:
                        # Accumulate gradients
                        if accum_grad is None:
                            accum_grad = grad
                        else:
                            accum_grad = tree_map(mx.add, grad, accum_grad)
                        mx.eval(accum_grad)
                        del grad

                        # Update model with accumulated gradients
                        if (n + 1) % gradient_accumulation_steps == 0:
                            accum_grad, _ = optim.clip_grad_norm(accum_grad, max_norm=gradient_max_norm)
                            accum_grad = tree_map(lambda g: g * accum_scale, accum_grad)
                            optimizer.update(self.model, accum_grad)
                            mx.eval(optimizer.state)
                            accum_grad = None
                    else:
                        # Update model with gradients
                        grad, _ = optim.clip_grad_norm(grad, max_norm=gradient_max_norm)
                        optimizer.update(self.model, grad)
                        mx.eval(optimizer.state)

                # Evaluate loss & reward
                mx.eval(loss_value, reward, num_tokens)

                # Record loss and number of tokens
                losses.append(loss_value.item())
                intv_tokens += num_tokens.item()

                # Record rewards
                if reward is not None:
                    reward = reward.tolist()

                    if rewards is None:
                        rewards = reward
                    else:
                        rewards = np.vstack([rewards, reward])

                # Report training loss if needed
                if (n + 1) % report_steps == 0:
                    train_loss = np.mean(losses)
                    stop = time.perf_counter()

                    # Print training loss and timings
                    pbar_epochs.write(f"#{n + 1}:\tTraining loss    {train_loss:.3f}\t{float(intv_tokens) / (stop - start):.3f} tok/sec (learning rate: {optimizer.learning_rate.item():.3e})")
                    if rewards is not None:
                        pbar_epochs.write(f"#{n + 1}:\tTraining reward  {str(np.mean(rewards, axis=0))}")
                        rewards = None

                    # Verbose logging using tqdm output
                    if logger.level <= logging.INFO:
                        # Print memory usage
                        peak_memory = mx.metal.get_peak_memory()
                        memory_usage = peak_memory / system_memory
                        pbar_epochs.write(f"#{n + 1}:\tPeak memory      {(peak_memory // (1024 ** 2)):,} MB ({memory_usage:.2%} of system memory)")
                        mx.metal.reset_peak_memory()

                    pbar_epochs.refresh()

                    if report_callback is not None:
                        report_callback(n + 1, train_loss)
                    
                    losses = []
                    intv_tokens = 0
                    start = time.perf_counter()

                # Report validation loss if needed
                if n == 0 or (n + 1) % eval_steps == 0:
                    # Print validation loss and timings
                    stop = time.perf_counter()
                    val_loss = self.evaluate(dataset_validation, batch_size, validation_batches)
                    start = time.perf_counter()
                    pbar_epochs.write(f"#{n + 1}:\tValidation loss  {val_loss:.3f}\t{(start - stop):.3f} sec")

                    # Eval callback
                    if eval_callback is not None:
                        msg = eval_callback(n + 1, val_loss)
                        if msg:
                            pbar_epochs.write(f"#{n + 1}:\t" + msg)

                    start = time.perf_counter()

        # Evaluate test dataset
        if dataset_test is not None:
            test_batches = validation_samples // batch_size

            stop = time.perf_counter()
            test_loss = self.evaluate(dataset_test, batch_size, test_batches)
            start = time.perf_counter()
            logger.info(f"Test loss: {test_loss:.3f}\t{start - stop:.3f} sec")


import os
import argparse
import readline

import sillm
import sillm.utils as utils

if __name__ == "__main__":
    # Parse commandline arguments
    parser = argparse.ArgumentParser(description="A simple CLI for generating text with SiLLM.")
    parser.add_argument("model", type=str, help="The model directory or file")
    parser.add_argument("-d", "--chdir", default=None, type=str, help="Change working directory")
    parser.add_argument("-c", "--config", default=None, type=str, help="Load YAML configuration file for chat")
    parser.add_argument("-a", "--input_adapters", default=None, type=str, help="Load and merge LoRA adapter weights from .safetensors file")
    parser.add_argument("-s", "--seed", type=int, default=-1, help="Seed for randomization")
    parser.add_argument("-t", "--temperature", type=float, default=0.7, help="Sampling temperature")
    parser.add_argument("-k", "--top_k", type=int, default=0, help="Top-k sampling")
    parser.add_argument("-p", "--top_p", type=float, default=1.0, help="Top-p sampling")
    parser.add_argument("-n", "--top_nsigma", type=float, default=0.0, help="Top-nsigma sampling")
    parser.add_argument("-r", "--repetition_penalty", type=float, default=None, help="Repetition penalty")
    parser.add_argument("-w", "--repetition_window", type=int, default=50, help="Window of generated tokens to consider for repetition penalty")
    parser.add_argument("-f", "--flush", type=int, default=5, help="Flush output every n tokens")
    parser.add_argument("-m", "--max_tokens", type=int, default=1024, help="Max. number of tokens to generate")
    parser.add_argument("-q", "--quantize", type=int, default=None, help="Quantize the model weights to the specified number of bits")
    parser.add_argument("--qkv", type=int, default=None, help="Quantize the KV cache to the specified number of bits")
    parser.add_argument("--cache", type=int, default=0, help="Create a prompt cache with the specified size")
    parser.add_argument("--template", type=str, default=None, help="Chat template (chatml, llama2, alpaca, etc.)")
    parser.add_argument("--system", type=str, default=None, help="System prompt for chat template")
    parser.add_argument("--ascii", default=False, action="store_true", help="Force output tokens to ASCII printable characters")
    parser.add_argument("-v", "--verbose", default=1, action="count", help="Increase output verbosity")
    args = parser.parse_args()

    # Change working directory
    if args.chdir is not None:
        os.chdir(args.chdir)

    # Load YAML configuration file
    if args.config is not None:
        utils.load_yaml(args.config, args)
    
    # Initialize logging
    log_level = 40 - (10 * args.verbose) if args.verbose > 0 else 0
    logger = utils.init_logger(log_level)

    # Log commandline arguments
    if log_level <= 10:
        utils.log_arguments(args.__dict__)

    # Set random seed
    if args.seed >= 0:
        utils.seed(args.seed)

    # Load model
    model = sillm.load(args.model)

    if args.input_adapters is not None:
        # Convert model to trainable
        model = sillm.TrainableLoRA.from_model(model)

        lora_config = model.load_lora_config(args.input_adapters)

        # Initialize LoRA layers
        model.init_lora(**lora_config)

        # Load and merge adapter file
        model.load_adapters(args.input_adapters)
        model.merge_and_unload_lora()

    # Quantize model
    if args.quantize is not None:
        model.quantize(bits=args.quantize)
    
    # Initialize prompt cache
    prompt_cache = None
    if args.cache > 0:
        prompt_cache = sillm.PromptCache(max_size=args.cache)

    # Initialize logit filters
    logit_filter = None
    if args.ascii:
        logit_filter = sillm.experimental.logit_filter.ASCIIFilter(model.tokenizer, model.args.vocab_size)

    generate_args = {
        "temperature": args.temperature,
        "top_k": args.top_k,
        "top_p": args.top_p,
        "top_nsigma": args.top_nsigma,
        "repetition_penalty": args.repetition_penalty,
        "repetition_window": args.repetition_window,
        "max_tokens": args.max_tokens,
        "flush": args.flush,
        "prompt_cache": prompt_cache,
        "logit_filter": logit_filter
    }

    kv_cache_args = {
        "quantized": args.qkv is not None,
        "bits": args.qkv
    }

    system_prompt = args.system
    if args.system is not None:
        if os.path.isfile(args.system):
            system_prompt = open(args.system, "r").read()

    # Initialize generator variables
    template = sillm.init_template(model.tokenizer, model.args, args.template)
    conversation = sillm.Conversation(template, system_prompt=system_prompt)
    cache = model.init_kv_cache(**kv_cache_args)

    # Log memory usage
    utils.log_memory_usage()

    # Input loop
    prompt = ""
    while True:
        prompt += input("> ")

        if prompt.startswith('/') or len(prompt) == 0:
            if prompt == "/exit":
                # Exit chat
                break
            elif prompt == "/clear":
                # Clear conversation
                conversation.clear()
                cache = model.init_kv_cache(**kv_cache_args)
            else:
                print("Commands:")
                print("/exit - Exit chat")
                print("/clear - Clear conversation")
            
            prompt = ""
            continue
        elif prompt.startswith('@'):
            # Load prompt from file
            fpath = prompt.lstrip('@')
            if not os.path.isfile(fpath):
                continue
            prompt = open(fpath, "r").read()
        elif prompt.endswith('\\'):
            # Continue prompt after line break
            prompt = prompt.rstrip('\\') + "\n"
            continue

        # Add user message to conversation and get prompt string
        request = conversation.add_user(prompt)

        logger.debug(f"Generating {args.max_tokens} tokens with temperature {args.temperature}")
        response = ""
        for s, metadata in model.generate(request, cache=cache, **generate_args):
            print(s, end="", flush=True)
            response += s
        print()

        conversation.add_assistant(response)
        prompt = ""

        logger.debug(f"Evaluated {metadata['usage']['prompt_tokens']} prompt tokens in {metadata['timing']['eval_time']:.2f}s ({metadata['usage']['prompt_tokens'] / metadata['timing']['eval_time']:.2f} tok/sec)")
        logger.debug(f"Generated {metadata['usage']['completion_tokens']} tokens in {metadata['timing']['runtime']:.2f}s ({metadata['usage']['completion_tokens'] / metadata['timing']['runtime']:.2f} tok/sec)")


import mlx.core as mx
import mlx.nn as nn

from sillm.models.base import BaseModel
from sillm.models.args import ModelArgs
import sillm.models.llama as llama
from sillm.modules.act import init_act

class NemotronLayerNorm1P(nn.LayerNorm):
    """
    Nemotron layer normalization.
    """
    def __call__(self, x):
        weight = self.weight + 1 if "weight" in self else None
        bias = self.bias if "bias" in self else None

        return mx.fast.layer_norm(x, weight, bias, self.eps)

class FeedForward(nn.Module):
    """
    Feed-forward module.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        super().__init__()

        self.w2 = nn.Linear(args.hidden_dim, args.dim, bias=False)
        self.w3 = nn.Linear(args.dim, args.hidden_dim, bias=False)
        
        self.act = init_act(args)
    
    def __call__(self,
                 x: mx.array
                 ) -> mx.array:
        """
        Args:
            x: Input tensor.
        Returns:
            Output tensor.
        """
        return self.w2(self.act(self.w3(x)))

class TransformerBlock(llama.TransformerBlock):
    """
    Transformer block.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        nn.Module.__init__(self)
        self.args = args

        self.n_heads = args.n_heads
        self.dim = args.dim

        self.attention = llama.Attention(args=args)
        self.feed_forward = FeedForward(args=args)
        self.attention_norm = NemotronLayerNorm1P(args.dim, eps=args.norm_eps)
        self.ffn_norm = NemotronLayerNorm1P(args.dim, eps=args.norm_eps)

class Model(llama.Model):
    """
    Nemotron model wrapper.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        BaseModel.__init__(self, args)
        self.args = args

        self.n_layers = args.n_layers
        self.vocab_size = args.vocab_size
        
        self.tok_embeddings = nn.Embedding(args.vocab_size, args.dim)
        self.layers = [TransformerBlock(args=args) for _ in range(args.n_layers)]
        self.norm = NemotronLayerNorm1P(args.dim, eps=args.norm_eps)

        if args.tie_word_embeddings:
            self.output = None
        else:
            self.output = nn.Linear(args.dim, args.vocab_size, bias=False)